unit AlgUtils;

{*******************************************************}
{                                                       }
{       Delphi Run-time Library                         }
{       Algorithm Utilities                             }
{                                                       }
{       2000..2004 YT                                   }
{                                                       }
{*******************************************************}

{$DEFINE SORTSWITCHING}

interface

uses
	Types, Math;

{ Math }

type
	Natural = 0..High(Integer); {R, Aday[Y̒`0}
	Radians = Extended;

function ArcAngle(const Y, X: Extended): Radians; {pxtZ܂}
function SubAngle(const A1, A2: Radians): Radians; {pxZ܂}

procedure AngleLength(const X, Y: Extended; var Angle, Length: Extended);

function ModuleDivide(M, D: Integer): Integer; overload;
function ModuleDivide(M, D: Int64): Int64; overload;
function ModuleRemainder(M, D: Integer): Integer; overload;
function ModuleRemainder(M, D: Int64): Int64; overload;

function GCD(A, B: Int64): Int64;

function AdjustRectMapping(const Area: TRect; var Dest, Source: TRect): Boolean;
//Dest  Source ւ̃}bvDestWAreaŐ
function Revise(Value, VMax, VMin, DeadZone, Range: Integer): Integer;
//͈͍ă}bv
function MinSuited(AtLeast, Now, New: Integer): Boolean;
//AtLeastȏōŒlIԂƂɎgNewNow葊

{ Color }

type
	HSV = record
		H, S, V: Extended;
	end;

function RGBToHSV(Value: Cardinal): HSV;
function HSVToRGB(const Value: HSV): Cardinal;

{ Hash }

type
	THashCode = Word; {nbVR[h}

function GetHashCode(const ABuffer; ACount: Integer): THashCode; register; overload;
function GetHashCode(const S: WideString): THashCode; overload;

{ String }

function IntToBin(Value: Integer): string;

procedure AddCharToUTF8String(var S: UTF8String; C: Char);
function QuoteAndEscape(const S: UTF8String): UTF8String;
function ExtractQuoteAndEscape(const S: UTF8String): UTF8String;

{ List }

procedure AddToArray(var PArray; var Count: Integer; Element: Pointer);
procedure RemoveFromArray(var PArray; var Count: Integer; Element: Pointer);

procedure AddToLinkList(var First, Last, This, Pred, Succ; const Next); overload;
procedure AddToLinkList(var First, Last, This, Pred, Succ); overload;
procedure RemoveFromLinkList(var First, Last, This, Pred, Succ);

procedure AddHistory(var A: TWideStringDynArray; const T: WideString; Limit: Integer); overload;
procedure AddFixedHistory(var A: array of WideString; const T: WideString); overload;

{ Sort }

const
	SortSwitchingBorder = 20; {}\[gɐ؂ւf[^}

type
	BComp = function(I1, I2: Integer; Context: Pointer): Integer; cdecl;
	BSwap = procedure(I1, I2: Integer; Context: Pointer); cdecl;

procedure InsertSort(Low, High: Integer; Comp: BComp; Swap: BSwap; Context: Pointer);
procedure QuickSort(Low, High: Integer; Comp: BComp; Swap: BSwap; Context: Pointer);

implementation

{ Math }

procedure AngleLength(const X, Y: Extended; var Angle, Length: Extended);
begin
	Length := Sqrt(Sqr(X) + Sqr(Y));
	Angle := ArcTan2(Y / Length, X / Length);
end;

function ArcAngle(const Y, X: Extended): Radians;
var
	L: Extended;
begin
	L := Sqrt(Sqr(X) + Sqr(Y));
	Result := ArcTan2(Y / L, X / L);
end;

function SubAngle(const A1, A2: Radians): Radians;
begin
	Result := A1 - A2;
	while Result > +PI do Result := Result - PI;
	while Result < -PI do Result := Result + PI;
end;

function ModuleDivide(M, D: Integer): Integer;
begin
	if Sign(M) <> Sign(D) then
		Result := (M - D + 1) div D
	else
		Result := M div D;
end;

function ModuleDivide(M, D: Int64): Int64; overload;
begin
	if Sign(M) <> Sign(D) then
		Result := (M - D + 1) div D
	else
		Result := M div D;
end;

function ModuleRemainder(M, D: Integer): Integer;
begin
	Result := M mod D;
	if Result < 0 then Inc(Result, D);
end;

function ModuleRemainder(M, D: Int64): Int64; overload;
begin
	Result := M mod D;
	if Result < 0 then Inc(Result, D);
end;

function GCD(A, B: Int64): Int64;
begin
	if B = 0 then
		Result := A
	else
		Result := GCD(B, A mod B)
end;

function AdjustRectMapping(const Area: TRect; var Dest, Source: TRect): Boolean;
var
	S: Integer;
begin
	if (Source.Bottom <= Source.Top) or (Source.Right <= Source.Left) or
					(Dest.Bottom <= Dest.Top) or (Dest.Right <= Dest.Left) or
					(Dest.Right < Area.Left) or (Dest.Left >= Area.Right) or
					(Dest.Bottom < Area.Top) or (Dest.Top >= Area.Bottom) then
	begin
		Result := False;
		Exit;
	end;

	S := Dest.Right - Area.Right;
	if S > 0 then
	begin
		Dec(Source.Right, (Source.Right - Source.Left) *
						S div (Dest.Right - Dest.Left));
		Dec(Dest.Right, S);
	end;

	S := Area.Left - Dest.Left;
	if S > 0 then
	begin
		Inc(Source.Left, (Source.Right - Source.Left) *
						S div (Dest.Right - Dest.Left));
		Inc(Dest.Left, S);
	end;

	S := Dest.Bottom - Area.Bottom;
	if S > 0 then
	begin
		Dec(Source.Bottom, (Source.Bottom - Source.Top) *
						S div (Dest.Bottom - Dest.Top));
		Dec(Dest.Bottom, S);
	end;

	S := Area.Top - Dest.Top;
	if S > 0 then
	begin
		Inc(Source.Top, (Source.Right - Source.Left) *
						S div (Dest.Bottom - Dest.Top));
		Inc(Dest.Top, S);
	end;

	Result := True;
end;

function Revise(Value, VMax, VMin, DeadZone, Range: Integer): Integer;
var
	OldRange, OldCenter, West, Sign: Integer;
begin
	OldRange  := (VMax - VMin) div 2;
	OldCenter := (VMax + VMin) div 2; Dec(Value, OldCenter); //S 0 

	West := OldRange * DeadZone div 100; //mȂ

	if (OldRange <= West) or (Abs(Value) <= West) then
		Result:= 0
	else
	begin
		if Value < 0 then
			Sign := -1
		else if Value > 0 then
			Sign := 1
		else
			Sign := 0;

		Result := (Value - Sign * West) * Range div (OldRange - West);
	end;
end;

function MinSuited(AtLeast, Now, New: Integer): Boolean;
begin
	if Now < AtLeast then
		Result := New > Now {vꂽɖȂꍇőI}
	else
		Result := (New >= AtLeast) and (New < Now);
		{vꂽ𒴂ꍇŏI}
end;

{ Color }

function RGBToHSV(Value: Cardinal): HSV;
var
	R, G, B, Max, Min, Diff: Integer;
begin
	R := Value and $ff;
	G := (Value shr 8) and $ff;
	B := (Value shr 16) and $ff;

	Max := R;
	if G > Max then Max := G;
	if B > Max then Max := B;
	Min := R;
	if G < Min then Min := G;
	if B < Min then Min := B;

	Diff := Max - Min;

	{x}
	Result.V := Max / 255;

	{ʓx}
	Result.S := 0;
	Result.H := 0;
	if Max > Min then
	begin
		Result.S := Diff / Max;

		{F}
		if B = Max then
			Result.H := 4 + (R - G) / Diff
		else if G = Max then
			Result.H := 2 + (B - R) / Diff
		else
		begin
			Result.H := (G - B) / Diff;
			if Result.H < 0 then
				Result.H := Result.H + 6;
		end;

		Result.H := Result.H * (2 * PI) / 6;
	end;
end;

function HSVToRGB(const Value: HSV): Cardinal;
var
	Diff: Extended;
	V, N1, N2, N3, R, G, B: Byte;
begin
	V := Trunc(Value.V * 255);

	if Value.S = 0 then
	begin
		Result := V or (V shl 8) or (V shl 16)
	end
	else
	begin
		Diff := Value.H * 6 / (2 * PI);
		Diff := Diff - Trunc(Diff);

		N1 := Trunc(Value.V * 255 * (1 - Value.S));
		N2 := Trunc(Value.V * 255 * (1 - Value.S * Diff));
		N3 := Trunc(Value.V * 255 * (1 - Value.S * (1 - Diff)));

		if      Value.H = 0          then begin R := V ; G := N1; B := N1 end
		else if Value.H < PI     / 3 then begin R := V ; G := N3; B := N1 end
		else if Value.H = PI     / 3 then begin R := V ; G := V ; B := N1 end
		else if Value.H < PI * 2 / 3 then begin R := N2; G := V ; B := N1 end
		else if Value.H = PI * 2 / 3 then begin R := N1; G := V ; B := N1 end
		else if Value.H < PI         then begin R := N1; G := V ; B := N3 end
		else if Value.H = PI         then begin R := N1; G := V ; B := V  end
		else if Value.H < PI * 4 / 3 then begin R := N1; G := N2; B := V  end
		else if Value.H = PI * 4 / 3 then begin R := N1; G := N1; B := V  end
		else if Value.H < PI * 5 / 3 then begin R := N3; G := N1; B := V  end
		else if Value.H = PI * 5 / 3 then begin R := V ; G := N1; B := V  end
		else                              begin R := V ; G := N1; B := N2 end
		;
		Result := R or (G shl 8) or (B shl 16)
	end
end;

{ Hash }

function GetHashCode(const ABuffer; ACount: Integer): THashCode; register;
asm
				{ ABuffer as EAX, ACount  as EDX }
				MOV     ECX,EDX
				MOV     EDX,EAX
				XOR     EAX,EAX
@@1:    ROL     AX,5
				XOR     AL,[EDX]
				INC     EDX
				DEC     ECX
				JNE     @@1
end;

function GetHashCode(const S: WideString): THashCode;
begin
	Result := GetHashCode(Pointer(S)^, Length(S) * SizeOf(WideChar))
end;

{ String }

function IntToBin(Value: Integer): string;
const
	B: array[0..1] of Char = ('0', '1');
begin
	Result := '';
	repeat
		Result := B[Value and 1] + Result;
		Value := Cardinal(Value) shr 1;
	until Value = 0;
end;

procedure AddCharToUTF8String(var S: UTF8String; C: Char);
var
	Len: Integer;
begin
	Len := Length(S);
	SetLength(S, Len + 1);
	PChar(Pointer(S))[Len] := C;
end;

function QuoteAndEscape(const S: UTF8String): UTF8String;
var
	I: Integer;
	Q: Boolean;
	N: string;
begin
	if S = '' then
		Result := ''''''
	else
	begin
		Result := '';
		Q := False;
		for I := 1 to Length(S) do
		begin
			case S[I] of
			#0..#31, '''':
			begin
				if Q then Result := Result + '''';
				Str(Ord(S[I]), N);
				Result := Result + '#' + N;
				Q := False;
			end;
			else
				if not Q then Result := Result + '''';
				Result := Result + S[I];
				Q := True;
			end;
		end;
		if Q then Result := Result + '''';
	end
end;

function ExtractQuoteAndEscape(const S: UTF8String): UTF8String;
var
	I, V, C, Len: Integer;
	Q: (Outside, CStyle, PascalStyle);
begin
	Result := '';
	Len := Length(S);
	Q := Outside;
	I := 1;
	while I <= Len do
	begin
		if S[I] = '''' then
		begin
			Inc(I);
			case Q of
			CStyle:      Result := Result + '''';
			PascalStyle: Q := Outside;
			else         Q := PascalStyle;
			end;
		end
		else if S[I] = '"' then
		begin
			Inc(I);
			case Q of
			CStyle:      Q := Outside;
			PascalStyle: Result := Result + '"';
			else         Q := CStyle;
			end;
		end
		else if (Q = Outside) and (S[I] = '#') then
		begin
			Inc(I);
			V := 0;
			while (I <= Len) and (S[I] in ['0'..'9']) do
			begin
				V := V * 10 + (Ord(S[I]) - Ord('0'));
				Inc(I);
			end;
			AddCharToUTF8String(Result, Chr(V));
		end
		else if (Q = CStyle) and (S[I] = '\') then
		begin
			Inc(I);
			if (I <= Len) and ((S[I] = 'x') or (S[I] = 'X')) then
			begin
				Inc(I);
			end;
			C := 0;
			V := 0;
			while I <= Len do
			begin
				if S[I] in ['a'..'f'] then
					C := Ord(S[I]) + (10 - Ord('a'))
				else if S[I] in ['A'..'F'] then
					C := Ord(S[I]) + (10 - Ord('A'))
				else if S[I] in ['0'..'9'] then
					C := Ord(S[I]) - Ord('0')
				else
					Break;

				V := V * 16 + C;
				Inc(I);
			end;
			AddCharToUTF8String(Result, Chr(V));
		end
		else
		begin
			AddCharToUTF8String(Result, S[I]);
			Inc(I);
		end;
	end;
end;

{ List }

procedure AddToArray(var PArray; var Count: Integer; Element: Pointer);
begin
	ReallocMem(Pointer(PArray), Count + 1);

	PPointerArray(PArray)^[Count] := Element;
	Inc(Count);
end;

procedure RemoveFromArray(var PArray; var Count: Integer; Element: Pointer);
var
	I: Integer;
	Src, Dst: ^Pointer;
begin
	for I:= 0 to Count -1 do
		if PPointerArray(PArray)^[I] = Element then
		begin
			Dec(Count);
			Dst:= Pointer(PArray); Inc(Dst, I); Src:= Dst; Inc(Src);
			Move(Src^, Dst^, (Count - I) * SizeOf(Pointer));

			Exit;
		end;

	Assert(False);
end;

procedure AddToLinkList(var First, Last, This, Pred, Succ; const Next); overload;
begin
	if Pointer(Next) = nil then
	begin
		Pointer(Pred) := Pointer(Last);
		Pointer(Succ) := nil
	end
	else
	begin
		Pointer(Pred) := PPointer(Cardinal(Next) + Cardinal(@Pred) - Cardinal(This))^;
		Pointer(Succ) := Pointer(Next);
	end;

	if Pointer(Pred) = nil then
		Pointer(First) := Pointer(This)
	else
		PPointer(Cardinal(Pred) + Cardinal(@Succ) - Cardinal(This))^ := Pointer(This);

	if Pointer(Succ) = nil then
		Pointer(Last) := Pointer(This)
	else
		PPointer(Cardinal(Succ) + Cardinal(@Pred) - Cardinal(This))^ := Pointer(This);
end;

procedure AddToLinkList(var First, Last, This, Pred, Succ); overload;
begin
	Pointer(Pred) := Pointer(Last);
	Pointer(Succ) := nil;

	if Pointer(Pred) = nil then
		Pointer(First) := Pointer(This)
	else
		PPointer(Cardinal(Pred) + Cardinal(@Succ) - Cardinal(This))^ := Pointer(This);

	Pointer(Last) := Pointer(This)
end;

procedure RemoveFromLinkList(var First, Last, This, Pred, Succ);
begin
	if Pointer(Pred) <> nil then
		PPointer(Cardinal(Pred) + Cardinal(@Succ) - Cardinal(This))^ := Pointer(Succ);

	if Pointer(Succ) <> nil then
		PPointer(Cardinal(Succ) + Cardinal(@Pred) - Cardinal(This))^ := Pointer(Pred);

	if Pointer(First) = Pointer(This) then
		Pointer(First) := Pointer(Succ);

	if Pointer(Last) = Pointer(This) then
		Pointer(Last) := Pointer(Pred);
end;

procedure AddHistory(var A: TWideStringDynArray; const T: WideString; Limit: Integer); overload;
var
	F, I: Integer;
begin
	F := Length(A);
	for I := High(A) downto Low(A) do
		if A[I] = T then
		begin
			F := I;
			Break;
		end;
	if F = Length(A) then SetLength(A, F + 1);
	for I := F downto 1 do A[I] := A[I - 1];
	A[0] := T;
	if Length(A) > Limit then SetLength(A, Limit);
end;

procedure AddFixedHistory(var A: array of WideString; const T: WideString);
var
	F, I: Integer;
begin
	F := High(A);
	for I := F - 1 downto Low(A) do
		if A[I] = T then
		begin
			F := I;
			Break;
		end;
	for I := F downto 1 do A[I] := A[I - 1];
	A[0] := T;
end;

{ Sort }

procedure InsertSort(Low, High: Integer; Comp: BComp; Swap: BSwap; Context: Pointer);
var
	I, J: Integer;
begin
	for I := Low + 1 to High do
		for J := I - 1 downto Low do
		begin
			if Comp(J, J +1, Context) < 0 then Break;
			Swap(J, J +1, Context);
		end;
end;

procedure QuickSort(Low, High: Integer; Comp: BComp; Swap: BSwap; Context: Pointer);
type
	TStackElement = record Low, High: Integer; end;
	TStack = array[0..32767] of TStackElement;
	PStack = ^TStack;
var
	I, J, Left, Right: Integer;
	s, Lo, Hi{, Temp}: Integer;
	sp: Integer;
	Stack: PStack;
begin
	GetMem(Stack, (High - Low + 2) * SizeOf(TStackElement));
	try

		Stack[0].Low  := Low;
		Stack[0].High := High;
		sp:= 1;

		while sp > 0 do
		begin
			Dec(sp);

			left := stack[sp].low;
			right:= stack[sp].high;

			if left < right then
			begin
{$ifdef SORTSWITCHING}
				if right - left > SortSwitchingBorder then
				begin
{$endif}
					// a[left],a[(left+right)/2],a[right]̒lƂ

					s := (left + right) div 2;
					lo:= left;
					hi:= right;

					if Comp(s, lo, Context) < 0 then
					begin
						{Temp:= s; s:= lo; lo:= Temp;} lo := s; s := left;
					end;

					if Comp(s, hi, Context) > 0 then
					begin
						{Temp:= s;} s:= hi; {hi:= Temp;}

						if Comp(s, lo, Context) < 0 then
						begin
							{Temp:= s;} s:= lo; {lo:= Temp;}
						end;
					end;

					i:= left  ;
					j:= right ;

					while True do
					begin
						while (i <> s) and (Comp(i, s, Context) < 0) do Inc(i);
						while (j <> s) and (Comp(j, s, Context) > 0) do Dec(j);

						if i >= j then
							Break;

						Swap(i, j, Context);

						//lł͂ȂQƂȂ̂ő[u
						if s = I then s:= J else if s = J then s:= I;

						Inc(I); Dec(J);
					end;

					// \[g͈͂X^bNɐς

					stack[sp].low  := j + 1;
					stack[sp].high := right;
					Inc(sp);

					stack[sp].low  := left;
					stack[sp].high := i - 1;
					Inc(sp);
{$ifdef SORTSWITCHING}
				end
				else
				begin
					// }\[g
					for I := Left + 1 to Right do
						for J := I - 1 downto Left do
						begin
							if Comp(J, J +1, Context) < 0 then Break;
							Swap(J, J +1, Context);
						end;
				end;
{$endif}
			end;
		end;

	finally
		FreeMem(Stack);
	end;
end;

end.
