unit Crt;

{*******************************************************}
{                                                       }
{       Windows CRT Interface Unit                      }
{                                                       }
{       WinCrt original interface                       }
{       Copytight (C) 1991,95 Borland International     }
{                                                       }
{       Win32 Console compatible and additional         }
{       2000..2003, 2007 YT                             }
{                                                       }
{*******************************************************}

{$WARN SYMBOL_PLATFORM OFF}

interface

uses
	Windows;

{ API }

const
	ATTACH_PARENT_PROCESS = DWORD(-1);

function GetConsoleWindow: THandle; stdcall; external kernel32;
function AttachConsole(ProcessId: DWORD): BOOL; stdcall; external kernel32;

{ WinCrt compatible }

const
	BLACK        = 0;
	BLUE         = FOREGROUND_BLUE;
	GREEN        = FOREGROUND_GREEN;
	CYAN         = FOREGROUND_GREEN or FOREGROUND_BLUE;
	RED          = FOREGROUND_RED;
	MAGENTA      = FOREGROUND_RED or FOREGROUND_BLUE;
	BROWN        = FOREGROUND_RED or FOREGROUND_GREEN;
	LIGHTGRAY    = FOREGROUND_RED or FOREGROUND_GREEN or FOREGROUND_BLUE;
	DARKGRAY     = FOREGROUND_INTENSITY;
	LIGHTBLUE    = FOREGROUND_INTENSITY or BLUE;
	LIGHTGREEN   = FOREGROUND_INTENSITY or GREEN;
	LIGHTCYAN    = FOREGROUND_INTENSITY or CYAN;
	LIGHTRED     = FOREGROUND_INTENSITY or RED;
	LIGHTMAGENTA = FOREGROUND_INTENSITY or MAGENTA;
	YELLOW       = FOREGROUND_INTENSITY or BROWN;
	WHITE        = FOREGROUND_INTENSITY or LIGHTGRAY;

procedure TextColor(Color: Cardinal);
procedure TextBackground(Color: Cardinal);
procedure GotoXY(X, Y: Integer);
procedure ClrScr;
procedure ClrEol;
function WhereX: Integer;
function WhereY: Integer;
procedure WriteBuf(Buffer: PChar; Count: Word);
procedure WriteChar(Ch: Char);
function ReadBuf(Buffer: PChar; Count: Word): Word;
function ReadKey: Char;
function KeyPressed: Boolean;

procedure AssignCrt(var F: Text); overload;

{ Input }

function CrtRead(var F: Text; B: PChar; L: Cardinal): Cardinal; overload;
function CrtRead(Input: THandle; B: PChar; L: Cardinal): Cardinal; overload;
function CrtReadChar(var F: Text): Char; overload;
function CrtReadChar(Input: THandle): Char; overload;
function CrtPeekChar(var F: Text): Char; overload;
function CrtPeekChar(Input: THandle): Char; overload;

function CrtHasEvents(var F: Text): Boolean; overload;
function CrtHasEvents(Input: THandle): Boolean; overload;
function CrtGetEvent(var F: Text; var Event: INPUT_RECORD): Boolean; overload;
function CrtGetEvent(Input: THandle; var Event: INPUT_RECORD): Boolean; overload;

{ TCrt }

type
	TCrt = record
	private
		function GetCursorPosition: TPoint;
		procedure SetCursorPosition(const Value: TPoint);
		function GetCursorVisible: Boolean;
		procedure SetCursorVisible(const Value: Boolean);
		function GetSize: TSize;
		procedure SetSize(const Value: TSize);
		function GetTextAttribute: Cardinal;
		procedure SetTextAttribute(const Value: Cardinal);
		class function GetTitle: string; static;
		class procedure SetTitle(const Value: string); static;
	public
		Handle: THandle;
		function Write(const Text: string): Cardinal; overload;
		function Write(T: PChar; L: Cardinal): Cardinal; overload;
		procedure WriteDirect(X, Y: Integer; const Text: String; Attribute: Cardinal); overload;
		procedure WriteDirect(X, Y: Integer; P: PChar; L: Cardinal; Attribute: Cardinal); overload;
		procedure WriteDirectTransparent(X, Y: Integer; const Text: String; ForeAttribute: Cardinal);
		procedure WriteDirectAttribute(const R: TRect; Attribute: Cardinal);
		procedure Clear;
		procedure ClearEoln;
		procedure Locate(X, Y: Integer);
		procedure Activate;
		property CursorPosition: TPoint read GetCursorPosition write SetCursorPosition;
		property CursorVisible: Boolean read GetCursorVisible write SetCursorVisible;
		property Size: TSize read GetSize write SetSize;
		property TextAttribute: Cardinal read GetTextAttribute write SetTextAttribute;
		class function IsForeground: Boolean; static;
		class property Title: string read GetTitle write SetTitle;
		class operator Explicit(const F: Text): TCrt;
	end;

function NewCrt: TCrt;
procedure DisposeCrt(Screen: TCrt);
procedure AssignCrt(var F: Text; Screen: TCrt); overload;
procedure AssignNewCrt(var F: Text);

{ Utilities }

function IsCrt: Boolean;

const
	ZeroCoord: COORD = (X: 0; Y: 0);
	RunUnderWin32Console = 'This program must be run under Win32 Console.';

implementation

{ WinCrt compatible }

procedure TextColor(Color: Cardinal);
var
	Screen: TCrt;
begin
	Screen := TCrt(Output);
	Screen.TextAttribute := Color or Screen.TextAttribute and not $f;
end;

procedure TextBackground(Color: Cardinal);
var
	Screen: TCrt;
begin
	Screen := TCrt(Output);
	Screen.TextAttribute := (Color shl 4) or Screen.TextAttribute and not $f0;
end;

procedure GotoXY(X, Y: Integer);
begin
  TCrt(Output).Locate(X - 1, Y - 1);
end;

function WhereX: Integer;
begin
	Result := TCrt(Output).CursorPosition.X + 1;
end;

function WhereY: Integer;
begin
	Result := TCrt(Output).CursorPosition.Y + 1;
end;

procedure ClrScr;
begin
	TCrt(Output).Clear
end;

procedure ClrEol;
begin
	TCrt(Output).ClearEoln
end;

procedure WriteBuf(Buffer: PChar; Count: Word);
begin
  TCrt(Output).Write(Buffer, Count)
end;

procedure WriteChar(Ch: Char);
begin
	TCrt(Output).Write(@Ch, 1)
end;

function ReadBuf(Buffer: PChar; Count: Word): Word;
begin
	Result := CrtRead(Input, Buffer, Count);
end;

function ReadKey: Char;
begin
	Result := CrtReadChar(Input);
end;

function KeyPressed: Boolean;
begin
	Result := CrtPeekChar(Input) <> #0;
end;

procedure AssignCrt(var F: Text);
begin
	Assign(F, '');
end;

{ Inputs }

function CrtRead(var F: Text; B: PChar; L: Cardinal): Cardinal; overload;
begin
	Result := CrtRead(TTextRec(F).Handle, B, L);
end;

function CrtRead(Input: THandle; B: PChar; L: Cardinal): Cardinal; overload;
begin
	ReadConsole(Input, B, L, Result, nil);
end;

function CrtReadChar(var F: Text): Char; overload;
begin
	Result := CrtReadChar(TTextRec(F).Handle);
end;

function CrtReadChar(Input: THandle): Char; overload;
var
	OrgMode, R: DWORD;
begin
	GetConsoleMode(Input, OrgMode);
	SetConsoleMode(Input, OrgMode and
					not (ENABLE_LINE_INPUT or ENABLE_ECHO_INPUT));
	ReadConsole(Input, @Result, 1, R, nil);
	SetConsoleMode(Input, OrgMode);
end;

function CrtPeekChar(var F: Text): Char; overload;
begin
	Result := CrtPeekChar(TTextRec(F).Handle);
end;

function CrtPeekChar(Input: THandle): Char; overload;
type
	TEventArray = array[0..32767] of INPUT_RECORD;
var
	PEvents: ^TEventArray;
	I, Count, Readed: Cardinal;
begin
	Result := #0;
	GetNumberOfConsoleInputEvents(Input, Count);
	if Count > 0 then
	begin
		GetMem(PEvents, SizeOf(INPUT_RECORD) * Count);
		try
			PeekConsoleInput(Input, PEvents^[0], Count, Readed);
			for I := 0 to Readed - 1 do
				if PEvents^[I].EventType = KEY_EVENT then
					if PEvents^[I].Event.KeyEvent.bKeyDown then
						if PEvents^[I].Event.KeyEvent.AsciiChar <> #0 then
						begin
							Result := PEvents^[I].Event.KeyEvent.AsciiChar;
							Exit;
						end;
		finally
			FreeMem(PEvents);
		end;
	end;
end;

function CrtHasEvents(var F: Text): Boolean;
begin
	Result := CrtHasEvents(TTextRec(F).Handle)
end;

function CrtHasEvents(Input: THandle): Boolean;
var
	E: DWORD;
begin
	GetNumberOfConsoleInputEvents(Input, E);
	Result := E > 0
end;

function CrtGetEvent(var F: Text; var Event: INPUT_RECORD): Boolean;
begin
	Result := CrtGetEvent(TTextRec(F).Handle, Event)
end;

function CrtGetEvent(Input: THandle; var Event: INPUT_RECORD): Boolean;
var
	Readed: DWORD;
begin
	Result := ReadConsoleInput(Input, Event, 1, Readed) and (Readed > 0)
end;

{ TCrt }

procedure TCrt.Activate;
begin
	SetConsoleActiveScreenBuffer(Handle)
end;

procedure TCrt.Clear;
const
	BufferSize: COORD = (X: $100; Y: $100);
var
	Buffer: array[0..$ff, 0..$ff] of CHAR_INFO;
	R: SMALL_RECT;
	S: TSize;
begin
	FillChar(Buffer, SizeOf(Buffer), 0);
	S := GetSize;
	R.Left := 0;
	R.Top := 0;
	R.Right := S.CX - 1;
	R.Bottom := S.CY - 1;
	WriteConsoleOutput(Handle, @Buffer, BufferSize, ZeroCoord, R);
	Locate(0, 0);
end;

procedure TCrt.ClearEoln;
const
	BufferSize: COORD = (X: $100; Y: 1);
var
	Buffer: array[0..$ff, 0..0] of CHAR_INFO;
	R: SMALL_RECT;
	P: TPoint;
	S: TSize;
begin
	FillChar(Buffer, SizeOf(Buffer), 0);
	P := GetCursorPosition;
	S := GetSize;
	R.Left := P.X;
	R.Top := P.Y;
	R.Right := S.CX - 1;
	R.Bottom := P.Y;
	WriteConsoleOutput(Handle, @Buffer, BufferSize, ZeroCoord, R);
end;

class operator TCrt.Explicit(const F: Text): TCrt;
begin
	if TTextRec(F).Mode = fmClosed then Rewrite(F);
	Result.Handle := TTextRec(F).Handle;
end;

function TCrt.GetCursorPosition: TPoint;
var
	Info: CONSOLE_SCREEN_BUFFER_INFO;
begin
	GetConsoleScreenBufferInfo(Handle, Info);
	Result.X := Info.dwCursorPosition.X;
	Result.Y := Info.dwCursorPosition.Y;
end;

function TCrt.GetCursorVisible: Boolean;
var
	Info: CONSOLE_CURSOR_INFO;
begin
	GetConsoleCursorInfo(Handle, Info);
	Result := Info.bVisible
end;

function TCrt.GetSize: TSize;
var
	Info: CONSOLE_SCREEN_BUFFER_INFO;
begin
	GetConsoleScreenBufferInfo(Handle, Info);
	Result.CX := Info.dwMaximumWindowSize.X;
	Result.CY := Info.dwMaximumWindowSize.Y;
end;

function TCrt.GetTextAttribute: Cardinal;
var
	Info: CONSOLE_SCREEN_BUFFER_INFO;
begin
	GetConsoleScreenBufferInfo(Handle, Info);
	Result := Info.wAttributes
end;

class function TCrt.GetTitle: string;
begin
	SetLength(Result, 1024);
	SetLength(Result, GetConsoleTitle(PChar(Result), Length(Result)) - 1)
end;

class function TCrt.IsForeground: Boolean;
begin
	Result := GetConsoleWindow = GetForegroundWindow
end;

procedure TCrt.Locate(X, Y: Integer);
var
	P: COORD;
begin
	P.X := X;
	P.Y := Y;
	SetConsoleCursorPosition(Handle, P);
end;

procedure TCrt.SetCursorPosition(const Value: TPoint);
begin
	Locate(Value.X, Value.Y)
end;

procedure TCrt.SetCursorVisible(const Value: Boolean);
var
	Info: CONSOLE_CURSOR_INFO;
begin
	GetConsoleCursorInfo(Handle, Info);
	Info.bVisible := Value;
	SetConsoleCursorInfo(Handle, Info);
end;

procedure TCrt.SetSize(const Value: TSize);
var
	T: COORD;
	Info: CONSOLE_SCREEN_BUFFER_INFO;
	R: SMALL_RECT;
begin
	T.X := Size.cx;
	T.Y := Size.cy;
	SetConsoleScreenBufferSize(Handle, T);
	GetConsoleScreenBufferInfo(Handle, Info);
	R.Left := 0;
	R.Top := 0;
	R.Right := Info.dwMaximumWindowSize.X - 1;
	R.Bottom := Info.dwMaximumWindowSize.Y - 1;
	SetConsoleWindowInfo(Handle, True, R);
end;

procedure TCrt.SetTextAttribute(const Value: Cardinal);
begin
	SetConsoleTextAttribute(Handle, Value);
end;

class procedure TCrt.SetTitle(const Value: string);
begin
	SetConsoleTitle(PChar(Value))
end;

function TCrt.Write(const Text: string): Cardinal;
begin
	WriteConsole(Handle, PChar(Text), Length(Text), Result, nil);
end;

function TCrt.Write(T: PChar; L: Cardinal): Cardinal;
begin
	WriteConsole(Handle, T, L, Result, nil);
end;

procedure TCrt.WriteDirect(X, Y: Integer; const Text: String; Attribute: Cardinal);
begin
  WriteDirect(X, Y, Pointer(Text), Length(Text), Attribute)
end;

procedure TCrt.WriteDirect(X, Y: Integer; P: PChar; L, Attribute: Cardinal);
var
	Buffer: array[0..0, 0..$ff] of CHAR_INFO;
	I: Integer;
	Size: COORD;
	R: SMALL_RECT;
begin
	L := L and $ff;
	for I := 0 to L - 1 do
	begin
		Buffer[0, I].AsciiChar := (P + I)^;
		Buffer[0, I].Attributes := Attribute;
	end;
	R.Left   := X;
	R.Top    := Y;
	R.Right  := X + Integer(L) - 1;
	R.Bottom := Y;
	Size.X := L;
	Size.Y := 1;
	WriteConsoleOutput(Handle, @Buffer, Size, ZeroCoord, R);
end;

procedure TCrt.WriteDirectAttribute(const R: TRect; Attribute: Cardinal);
var
	X, Y, Len: Integer;
	Attrs: array[0..159] of Word;
	S: COORD;
	Ret: DWORD;
begin
	if R.Top > R.Bottom then Exit;
	Len := R.Right - R.Left + 1;
	if Len <= 0 then Exit;
	if R.Left + Len > 80 then Len := 80 - R.Left;
	for X := Len - 1 downto 0 do
	begin
		Attrs[X]:= Attribute;
	end;
	S.X := R.Left;
	for Y := R.Top to R.Bottom do
	begin
		S.Y := Y;
		WriteConsoleOutputAttribute(Handle, @Attrs, Len, S, Ret);
	end;
end;

procedure TCrt.WriteDirectTransparent(X, Y: Integer; const Text: String; ForeAttribute: Cardinal);
var
	Buffer: array[0..0, 0..159] of CHAR_INFO;
	L, I: Integer;
	Size: COORD;
	R: SMALL_RECT;
begin
	L := Length(Text);
	Assert(L <= 160);
	R.Left := X; R.Top := Y; R.Right := X + L - 1; R.Bottom := Y;
	Size.X := L; Size.Y := 1;
	ReadConsoleOutput(Handle, @Buffer, Size, ZeroCoord, R);
	for I := 0 to L - 1 do
	begin
		Buffer[0, I].AsciiChar := (PChar(Text) + I)^;
		Buffer[0, I].Attributes := Buffer[0, I].Attributes and $f0 or ForeAttribute;
	end;
	WriteConsoleOutput(Handle, @Buffer, Size, ZeroCoord, R);
end;

function NewCrt: TCrt;
begin
	Result.Handle := CreateConsoleScreenBuffer(
		GENERIC_READ or GENERIC_WRITE, 0, nil, CONSOLE_TEXTMODE_BUFFER, nil)
end;

procedure DisposeCrt(Screen: TCrt);
begin
	CloseHandle(Screen.Handle)
end;	

procedure AssignCrt(var F: Text; Screen: TCrt); overload;

	function NoAction(var T: Text): Integer;
	begin
		Result := 0;
	end;

begin
	Assign(F, '');
	Rewrite(F);
	TTextRec(F).Handle := Screen.Handle;
	TTextRec(F).OpenFunc := @NoAction;
	TTextRec(F).CloseFunc := @NoAction;
end;

procedure AssignNewCrt(var F: Text);

	function NoAction(var T: Text): Integer;
	begin
		Result := 0;
	end;

	function OnClose(var T: Text): Integer;
	begin
		CloseHandle(TTextRec(T).Handle);
		Result := 0;
	end;

begin
	Assign(F, '');
	Rewrite(F);
	TTextRec(F).Handle := NewCrt.Handle;
	TTextRec(F).OpenFunc := @NoAction;
	TTextRec(F).CloseFunc := @OnClose;
end;

procedure PushAnyKey;
const
	Msg = '[ Please press any key to terminate ]';
var
	Input, Output: THandle;
	Done: DWORD;
	Cr: CONSOLE_CURSOR_INFO;
	InpRec: INPUT_RECORD;
begin
	Output := GetStdHandle(STD_OUTPUT_HANDLE);
	SetConsoleActiveScreenBuffer(Output);
	Cr.dwSize := 1;
	Cr.bVisible := False;
	SetConsoleCursorInfo(Output, Cr);

	SetConsoleTextAttribute(Output, 7 or BACKGROUND_RED);
	WriteConsole(Output, PChar(Msg), Length(Msg), Done, nil);

	Input := GetStdHandle(STD_INPUT_HANDLE);
	SetConsoleMode(Input, 0);

	FlushConsoleInputBuffer(Input);
	repeat
		ReadConsoleInput(Input, InpRec, 1, Done);
	until (Done = 1) and (InpRec.EventType = KEY_EVENT) and (InpRec.Event.KeyEvent.bKeyDown);
end;

{ Utilities }

function IsCrt: Boolean;
var
	Mode: DWORD;
	Info: CONSOLE_SCREEN_BUFFER_INFO;
begin
	Result := GetConsoleMode(GetStdHandle(STD_INPUT_HANDLE), Mode) and
		GetConsoleScreenBufferInfo(GetStdHandle(STD_OUTPUT_HANDLE), Info);
end;

initialization

finalization

	if (DebugHook <> 0) and IsCrt then PushAnyKey

end.

