unit ObjRAII;

{*******************************************************}
{                                                       }
{       Delphi Run-time Library                         }
{       Auto initialization/finalization                }
{                                                       }
{       2008 YT                                         }
{       Original idea: http://twc.xrea.jp/?20080125     }
{                                                       }
{*******************************************************}

interface

type
	TControlledTable = type string;

	PControlledTypeInfo = ^TControlledTypeInfo;
	TControlledTypeInfo = packed record
		Kind: Byte;
		Name: string[10]; { 2 + 4 + 4 }
		VMT: Pointer;
		SelfInfo: Pointer;
	end;

	PControlled = ^TControlled;
	TControlled = object
	protected
		procedure Initialize; virtual;
		procedure Finalize; virtual;
		procedure Adjust; virtual;
	public
		constructor Register(TypeInfo: Pointer);
	end;

implementation

uses
	SysHacks, Windows;

const
	tkControlled = 26; { Ord(High(TTypeKind)) + 9; }

{ Runtime patch }

procedure ReplaceProc(OldProc, NewProc: Pointer);
var
	OldProtect, Dummy: Cardinal;
begin
	VirtualProtect(OldProc, 5, PAGE_READWRITE, @OldProtect);
	PByte(OldProc)^ := $e9; {jmp}
	PInteger(Integer(OldProc) + 1)^ := Integer(NewProc) - Integer(OldProc) - 5;
	VirtualProtect(OldProc, 5, OldProtect, @Dummy);
end;

procedure ReplacePointer(var Old; const New: Pointer);
var
	OldProtect, Dummy: Cardinal;
begin
	VirtualProtect(@Old, SizeOf(Pointer), PAGE_READWRITE, @OldProtect);
	Pointer(Old) := New;
	VirtualProtect(@Old, SizeOf(Pointer), OldProtect, @Dummy);
end;

{ System unit replacing }

procedure _InitializeArray(p: Pointer; typeInfo: Pointer; elemCount: Cardinal);
var
	FT: PFieldTable;
	Self: PControlled;
begin
	if elemCount = 0 then Exit;
	case PTypeInfo(typeInfo).Kind of
		tkLString, tkWString, tkInterface, tkDynArray:
			while elemCount > 0 do
			begin
				PInteger(P)^ := 0;
				Inc(Integer(P), 4);
				Dec(elemCount);
			end;
		tkVariant:
			while elemCount > 0 do
			begin
				PInteger(P)^ := 0;
				PInteger(Integer(P)+4)^ := 0;
				PInteger(Integer(P)+8)^ := 0;
				PInteger(Integer(P)+12)^ := 0;
				Inc(Integer(P), sizeof(TVarData));
				Dec(elemCount);
			end;
		tkArray:
			begin
				FT := PFieldTable(Integer(typeInfo) + Byte(PTypeInfo(typeInfo).Name[0]));
				while elemCount > 0 do
				begin
					ObjRAII._InitializeArray(P, FT.Fields[0].TypeInfo^, FT.Count);
					Inc(Integer(P), FT.Size);
					Dec(elemCount);
				end;
			end;
		tkRecord:
			begin
				FT := PFieldTable(Integer(typeInfo) + Byte(PTypeInfo(typeInfo).Name[0]));
				while elemCount > 0 do
				begin
					SysHacks.InitializeRecord(P, typeInfo);
					Inc(Integer(P), FT.Size);
					Dec(elemCount);
				end;
			end;
		tkControlled:
			begin
				Self := PControlled(Integer(P) - SizeOf(Pointer));
				PPointer(Self)^ := PControlledTypeInfo(typeInfo).VMT;
				Self^.Initialize;
      end;	
  else
		System.Error(reInvalidPtr);
	end;
end;

procedure _FinalizeArray(p: Pointer; typeInfo: Pointer; elemCount: Cardinal);
var
	FT: PFieldTable;
	Self: PControlled;
begin
  if elemCount = 0 then Exit;
  case PTypeInfo(typeInfo).Kind of
		tkLString: _LStrArrayClr(P^, elemCount);
		tkWString: _WStrArrayClr(P^, elemCount);
		tkVariant:
			while elemCount > 0 do
			begin
				SysHacks._VarClr(PVarData(P)^);
				Inc(Integer(P), sizeof(TVarData));
				Dec(elemCount);
			end;
		tkArray:
			begin
				FT := PFieldTable(Integer(typeInfo) + Byte(PTypeInfo(typeInfo).Name[0]));
				while elemCount > 0 do
				begin
					ObjRAII._FinalizeArray(P, FT.Fields[0].TypeInfo^, FT.Count);
					Inc(Integer(P), FT.Size);
					Dec(elemCount);
				end;
			end;
		tkRecord:
			begin
				FT := PFieldTable(Integer(typeInfo) + Byte(PTypeInfo(typeInfo).Name[0]));
				while elemCount > 0 do
				begin
					SysHacks.FinalizeRecord(P, typeInfo);
					Inc(Integer(P), FT.Size);
					Dec(elemCount);
				end;
			end;
		tkInterface:
			while elemCount > 0 do
			begin
				SysHacks._IntfClear(IInterface(P^));
				Inc(Integer(P), 4);
				Dec(elemCount);
      end;
    tkDynArray:
			while elemCount > 0 do
			begin
				System.DynArrayClear(P, typeInfo);
        Inc(Integer(P), 4);
				Dec(elemCount);
			end;
		tkControlled:
			begin
				Self := PControlled(Integer(P) - SizeOf(Pointer));
				Self^.Finalize;
			end;
  else
		System.Error(reInvalidPtr);
  end;
end;

procedure _CopyRecord{ dest, source, typeInfo: Pointer };
asm
        { ->    EAX pointer to dest             }
				{       EDX pointer to source           }
				{       ECX pointer to typeInfo         }

				PUSH    EBX
				PUSH    ESI
				PUSH    EDI
				PUSH    EBP
	push 0

				MOV     EBX,EAX
				MOV     ESI,EDX

				XOR     EAX,EAX
				MOV     AL,[ECX+1]

				LEA     EDI,[ECX+EAX+2+8]
				MOV     EBP,[EDI-4]
				XOR     EAX,EAX
				MOV     ECX,[EDI-8]
				PUSH    ECX
@@loop:
				MOV     ECX,[EDI+4]
        SUB     ECX,EAX
        JLE     @@nomove1
        MOV     EDX,EAX
        ADD     EAX,ESI
        ADD     EDX,EBX
        CALL    System.Move
@@noMove1:
        MOV     EAX,[EDI+4]

        MOV     EDX,[EDI]
        MOV     EDX,[EDX]
        MOV     CL,[EDX]

        CMP     CL,tkLString
        JE      @@LString
        CMP     CL,tkWString
        JE      @@WString
        CMP     CL,tkVariant
        JE      @@Variant
        CMP     CL,tkArray
        JE      @@Array
				CMP     CL,tkRecord
				JE      @@Record
				CMP     CL,tkInterface
        JE      @@Interface
        CMP     CL,tkDynArray
				JE      @@DynArray
	cmp cl, tkControlled
	je @@Controlled
        MOV     AL,reInvalidPtr
        POP     EBP
        POP     EDI
        POP     ESI
        POP     EBX
				JMP     System.Error

@@LString:
        MOV     EDX,[ESI+EAX]
        ADD     EAX,EBX
				CALL    System.@LStrAsg
        MOV     EAX,4
        JMP     @@common

@@WString:
        MOV     EDX,[ESI+EAX]
        ADD     EAX,EBX
				CALL    System.@WStrAsg
        MOV     EAX,4
        JMP     @@common

@@Variant:
        LEA     EDX,[ESI+EAX]
        ADD     EAX,EBX
        CALL    System._VarCopy
        MOV     EAX,16
        JMP     @@common

@@Array:
        XOR     ECX,ECX
				MOV     CL,[EDX+1]
        PUSH    dword ptr [EDX+ECX+2]
        PUSH    dword ptr [EDX+ECX+2+4]
        MOV     ECX,[EDX+ECX+2+8]
        MOV     ECX,[ECX]
        LEA     EDX,[ESI+EAX]
        ADD     EAX,EBX
        CALL    System.@CopyArray
        POP     EAX
        JMP     @@common

@@Record:
        XOR     ECX,ECX
        MOV     CL,[EDX+1]
        MOV     ECX,[EDX+ECX+2]
        PUSH    ECX
        MOV     ECX,EDX
        LEA     EDX,[ESI+EAX]
        ADD     EAX,EBX
        CALL    ObjRAII._CopyRecord
        POP     EAX
        JMP     @@common

@@Interface:
				MOV     EDX,[ESI+EAX]
				ADD     EAX,EBX
				CALL    System.@IntfCopy
				MOV     EAX,4
				JMP     @@common

@@Controlled:
	mov edx, [ebx]
	mov [esp + 4], 1
	mov edx, [edx + vmtoffset TControlled.Finalize]
	mov eax, ebx
	call edx
	mov eax, 4
	jmp @@common

@@DynArray:
				MOV     ECX,EDX
				MOV     EDX,[ESI+EAX]
				ADD     EAX,EBX
				CALL    System.@DynArrayAsg
				MOV     EAX,4

@@common:
				ADD     EAX,[EDI+4]
				ADD     EDI,8
				DEC     EBP
				JNZ     @@loop

				POP     ECX
				SUB     ECX,EAX
				JLE     @@noMove2
				LEA     EDX,[EBX+EAX]
				ADD     EAX,ESI
				CALL    System.Move
@@noMove2:

	cmp [esp + 4], 0
	jz @@exit
	mov edx, [ebx]
	mov edx, [edx + vmtoffset TControlled.Adjust]
	mov eax, ebx
	call edx
@@exit:
	pop eax

				POP     EBP
				POP     EDI
				POP     ESI
				POP     EBX
end;

{ TControlled }

procedure TControlled.Adjust;
begin
end;

procedure TControlled.Finalize;
begin
end;

procedure TControlled.Initialize;
begin
end;

constructor TControlled.Register(TypeInfo: Pointer);
var
	VMT: Pointer;
	FT: PFieldTable;
	CT_TypeInfo: SysHacks.PTypeInfo;
	NewTypeInfo: PControlledTypeInfo;
begin
	{ eax <- Self(nil) }
	{ edx <- VMT }
	{ ecx <- TypeInfo }
	VMT := PPointer(@Self)^;

	//WriteLn(Cardinal(VMT));

	Dispose(@Self);

	Assert(PTypeInfo(TypeInfo).Kind = tkRecord);

	//WriteLn(Cardinal(TypeInfo));

	FT := PFieldTable(Integer(TypeInfo) + Byte(PTypeInfo(TypeInfo).Name[0]));

	Assert(FT.Count > 0);

	CT_TypeInfo := System.TypeInfo(TControlledTable);

	//WriteLn(CT_TypeInfo.Name);
	//WriteLn(FT.Fields[0].TypeInfo^.Name);

	Assert(FT.Fields[0].TypeInfo^ = CT_TypeInfo);
	Assert(FT.Fields[0].Offset = SizeOf(Pointer));

	New(NewTypeInfo);
	RegisterExpectedMemoryLeak(NewTypeInfo);

	NewTypeInfo^.Kind := tkControlled;
	NewTypeInfo^.Name := 'Controlled'; { 2 + 4 + 4 }
	NewTypeInfo^.VMT := VMT;
	NewTypeInfo^.SelfInfo := NewTypeInfo;

	ReplacePointer(FT.Fields[0].TypeInfo, @NewTypeInfo^.SelfInfo);
end;

initialization

	asm
		mov eax, offset System.@InitializeArray
		mov edx, offset ObjRAII._InitializeArray
		call ReplaceProc
		mov eax, offset System.@FinalizeArray
		mov edx, offset ObjRAII._FinalizeArray
		call ReplaceProc
		mov eax, offset System.@CopyRecord
		mov edx, offset ObjRAII._CopyRecord
		call ReplaceProc
	end

end.
