unit SysHacks;

{*******************************************************}
{                                                       }
{       Delphi Run-time Library                         }
{       Tricks                                          }
{                                                       }
{       2000..2004, 2007..2008 YT                       }
{                                                       }
{*******************************************************}

interface

{ Downward closure }

function GetContext: Pointer; register;

{ C# or Ruby like iterator }

type
	TEnumerator = class;
	TEnumeratorClass = class of TEnumerator;
	TEnumerator = class(TObject)
	private
		caller_ebx, caller_esi, caller_edi, caller_ebp, caller_eip: Pointer;
		enum_ebx, enum_esi, enum_edi, enum_esp, enum_eip: Pointer;
		save_esp, save_eip: Pointer;
		base_esp: Pointer;
		FNext, FCurrent: Pointer;
		FState: (esStart, esEnumerating, esTerminated, esEnded);
	public
		procedure BeforeDestruction; override;
		function MoveNext: Boolean; register;
		property GetCurrent: Pointer read FCurrent;
		class function Each(Item: Pointer; Self: TEnumerator): Boolean cdecl; static;
		class function GetEnumerator: TEnumerator; register;
	end;

(* example

{ Callback routine should have this signature. X is any register-size data. }
type
	CEach = function(X: Integer; Context: Pointer): Boolean; cdecl;

{ Iterator should NOT install exception handler. }
{ Iterator should make stack frame. }
procedure EnumTest(Proc: CEach; Context: Pointer);
begin
	if Proc(1, Context) then Exit;
	if Proc(2, Context) then Exit;
	if Proc(3, Context) then Exit;
end;

type
	TTestEnumerator = class(TEnumerator)
		function GetCurrent: Integer;
		property Current: Integer read GetCurrent;
	end;
	TTestCollection = class
		function GetEnumerator: TTestEnumerator;
	end;

function TTestEnumerator.GetCurrent: Integer;
begin
	Result := Integer(inherited GetCurrent)
end;

{ GetEnumerator should make stack frame. }
function TTestCollection.GetEnumerator: TTestEnumerator;
	procedure Call;
	begin
		EnumTest(CEach(@TTestEnumerator.Each), Result);
	end;
begin
	Result := TTestEnumerator(TTestEnumerator.GetEnumerator);
	Call;
end;

*)

{ Run time type info }

const
  tkLString   = 10;
  tkWString   = 11;
  tkVariant   = 12;
  tkArray     = 13;
  tkRecord    = 14;
  tkInterface = 15;
  tkDynArray  = 17;

type
  PPTypeInfo = ^PTypeInfo;
  PTypeInfo = ^TTypeInfo;
  TTypeInfo = packed record
    Kind: Byte;
    Name: ShortString;
   {TypeData: TTypeData}
  end;

  TFieldInfo = packed record
    TypeInfo: PPTypeInfo;
    Offset: Cardinal;
	end;

	PFieldTable = ^TFieldTable;
	TFieldTable = packed record
		X: Word;
		Size: Cardinal;
		Count: Cardinal;
		Fields: array [0..0] of TFieldInfo;
	end;

{ Polymorphic record }

function _New(size: Longint; typeInfo: Pointer): Pointer;
procedure _Dispose(p: Pointer; typeInfo: Pointer);

procedure InitializeRecord(p: Pointer; typeInfo: Pointer);
procedure FinalizeRecord(P: Pointer; typeInfo: Pointer);

procedure CopyRecord(Dest, Source, TypeInfo: Pointer);
procedure AddRefRecord(p: Pointer; typeInfo: Pointer);

{ Polymorphic array }

procedure InitializeArray(p: Pointer; typeInfo: Pointer; elemCount: Cardinal);
procedure FinalizeArray(p: Pointer; typeInfo: Pointer; elemCount: Cardinal);

procedure CopyArray(Dest, Source, TypeInfo: Pointer; Count: Integer);
procedure AddRefArray(p: Pointer; typeInfo: Pointer; elemCount: Longint);

procedure _LStrArrayClr(var StrArray; cnt: longint);
procedure _WStrArrayClr(var StrArray; Count: Integer);

{ Polymorphic dynamic array }

function DynArrayLength(A: Pointer): Integer;
procedure DynArrayCopy(a: Pointer; typeInfo: Pointer; var Result: Pointer);
procedure DynArrayCopyRange(a: Pointer; typeInfo: Pointer; index, count : Integer; var Result: Pointer);
procedure DynArrayAsg(var A: Pointer; Source: Pointer; TypeInfo: PDynArrayTypeInfo);
procedure DynArrayAddRef(A: Pointer);

(* System.pas
procedure DynArrayClear(var a: Pointer; typeInfo: Pointer);
procedure DynArraySetLength(var a: Pointer; typeInfo: Pointer; dimCnt: Longint; lengthVec: PLongint);
function DynArrayDim(typeInfo: PDynArrayTypeInfo): Integer;
*)

procedure DynArrayUnique(var A: Pointer; TypeInfo: PDynArrayTypeInfo);

{ Variants }

procedure _VarClr(var v: TVarData);
procedure _VarCopy(var Dest: TVarData; const Src: TVarData);

{ Interfaces }

function _IntfClear(var Dest: IInterface): Pointer;

implementation

{ Downward closure }

function GetContext: Pointer; register;
asm
	mov eax, ebp
end;

{ C# or Ruby like iterator }

class function TEnumerator.Each(Item: Pointer; Self: TEnumerator): Boolean; cdecl;
asm
	mov edx, Item
	mov eax, Self
	push $12345678 //ecx {?... dummy, for escaping crash when exception }
	push ebp
	mov TEnumerator[eax].FNext, edx
	mov TEnumerator[eax].enum_ebx, ebx
	mov TEnumerator[eax].enum_esi, esi
	mov TEnumerator[eax].enum_edi, edi
	mov TEnumerator[eax].enum_esp, esp
	mov TEnumerator[eax].enum_eip, offset @Next
	mov ebx, TEnumerator[eax].caller_ebx
	mov esi, TEnumerator[eax].caller_esi
	mov edi, TEnumerator[eax].caller_edi
	mov edx, TEnumerator[eax].caller_eip
	mov ebp, TEnumerator[eax].caller_ebp
	jmp edx
@Next:
	pop ebp
	pop ecx {?... pop dummy}
	mov eax, Self
	mov al, TEnumerator[eax].FState
	cmp al, esTerminated
	setae al
end;

class function TEnumerator.GetEnumerator: TEnumerator; register;
asm
	{CX^X쐬, eax = VMT}
	mov dl, 1
	call TEnumerator.Create
	lea edx, [ebp + 8] {for..inesp, ĂяoƂX^bNt[̑O}
	mov TEnumerator[eax].base_esp, edx
	mov TEnumerator[eax].caller_ebx, ebx
	mov TEnumerator[eax].caller_esi, esi
	mov TEnumerator[eax].caller_edi, edi
	mov edx, [ebp]
	mov TEnumerator[eax].caller_ebp, edx
	mov edx, [ebp + 4] {for..iñ^[AhX𓾂}
	mov TEnumerator[eax].caller_eip, edx
	{ĂяoƂ̌ケɖ߂Ă悤}
	mov [ebp + 4], offset @Return
	ret
@Return:
	{[vɏIĖ߂ĂȂ璆fʒu}
	mov dl, TEnumerator[eax].FState
	mov TEnumerator[eax].FState, esEnded
	test dl, dl
	jz @Exit
	mov ebx, TEnumerator[eax].caller_ebx
	mov esi, TEnumerator[eax].caller_esi
	mov edi, TEnumerator[eax].caller_edi
	mov ebp, TEnumerator[eax].caller_ebp
@Exit:
	mov edx, TEnumerator[eax].caller_eip
	jmp edx
end;

procedure TEnumerator.BeforeDestruction;
asm
	mov edx, TEnumerator[eax].base_esp
	sub edx, $400 {Oɂ銪߂̓X^bNRgĂ̂ŋʂł}
	cmp esp, edx
	jb @Exit
	mov dl, TEnumerator[eax].FState
	cmp dl, esEnumerating
	jnz @Exit
	{ItO}
	mov TEnumerator[eax].FState, esTerminated
	{ReLXgۑ}
	pop ecx
	mov TEnumerator[eax].save_eip, ecx
	mov TEnumerator[eax].save_esp, esp
	{XCb`}
	mov TEnumerator[eax].caller_ebx, ebx
	mov TEnumerator[eax].caller_esi, esi
	mov TEnumerator[eax].caller_edi, edi
	mov TEnumerator[eax].caller_ebp, ebp
	mov TEnumerator[eax].caller_eip, offset @Next
	mov ebx, TEnumerator[eax].enum_ebx
	mov esi, TEnumerator[eax].enum_esi
	mov edi, TEnumerator[eax].enum_edi
	mov esp, TEnumerator[eax].enum_esp
	mov edx, TEnumerator[eax].enum_eip
	jmp edx
@Next:
	{ReLXgA}
	mov esp, TEnumerator[eax].save_esp
	mov ecx, TEnumerator[eax].save_eip
	push ecx
	{X^bN߂}
	mov eax, TEnumerator[eax].base_esp
	lea edx, [eax - 24]
	mov eax, [esp] {return to BeforeDestruction}
	mov [edx], eax
	mov eax, [esp + 4] {pop in BeforeDestruction}
	mov [edx + 4], eax
	mov eax, [esp + 8] {pop in BeforeDestruction}
	mov [edx + 8], eax
	mov eax, [esp + 12] {return to Destroy}
	mov [edx + 12], eax
	mov eax, [esp + 16] {return to finally block}
	mov [edx + 16], eax
	mov eax, [esp + 20] {return to normal context}
	mov [edx + 20], eax
	mov esp, edx
@Exit:
end;

function TEnumerator.MoveNext: Boolean;
asm
	{FCurrent := FNext}
	mov edx, TEnumerator[eax].FNext
	mov TEnumerator[eax].FCurrent, edx
	{I}
	xor edx, edx
	mov dl, TEnumerator[eax].FState
	mov edx, dword ptr[@JumpTable + edx * 4]
	jmp edx
@Start:
	mov TEnumerator[eax].FState, esEnumerating
@Enumerating:
	{ReLXgۑ}
	pop ecx
	mov TEnumerator[eax].save_esp, esp
	mov TEnumerator[eax].save_eip, ecx
	{XCb`}
	mov TEnumerator[eax].caller_ebx, ebx
	mov TEnumerator[eax].caller_esi, esi
	mov TEnumerator[eax].caller_edi, edi
	mov TEnumerator[eax].caller_ebp, ebp
	mov TEnumerator[eax].caller_eip, offset @Next
	mov ebx, TEnumerator[eax].enum_ebx
	mov esi, TEnumerator[eax].enum_esi
	mov edi, TEnumerator[eax].enum_edi
	mov esp, TEnumerator[eax].enum_esp
	mov edx, TEnumerator[eax].enum_eip
	jmp edx
@Next:
	{ReLXgA}
	mov esp, TEnumerator[eax].save_esp
	mov ecx, TEnumerator[eax].save_eip
	push ecx
	mov dl, TEnumerator[eax].FState
	cmp dl, esEnded
	jne @Continue
	{X^bN߂}
	mov eax, TEnumerator[eax].base_esp
	lea edx, [eax - 16] {^[AhX, Onhpush*3}
	mov eax, esp
	mov ecx, 16
	push edx
	call System.Move
	pop esp
@Continue:
	mov al, 1
	ret
@Terminated:
	mov TEnumerator[eax].FState, esEnded
	mov al, 1
	ret
@Ended:
	xor al, al
	ret
@JumpTable:
	dd @Start
	dd @Enumerating
	dd @Terminated
	dd @Ended
end;

{ Polymorphic record }

function _New(size: Longint; typeInfo: Pointer): Pointer;
asm
	jmp System.@New
end;

procedure _Dispose(p: Pointer; typeInfo: Pointer);
asm
	jmp System.@Dispose
end;

procedure InitializeRecord(p: Pointer; typeInfo: Pointer);
asm
	jmp System.@InitializeRecord
end;

procedure FinalizeRecord(P: Pointer; typeInfo: Pointer);
asm
	jmp System.@FinalizeRecord
end;

procedure CopyRecord(Dest, Source, TypeInfo: Pointer);
asm
	jmp System.@CopyRecord
end;

procedure AddRefRecord(p: Pointer; typeInfo: Pointer);
asm
	jmp System.@AddRefRecord
end;

{ Polymorphic array }

procedure InitializeArray(p: Pointer; typeInfo: Pointer; elemCount: Cardinal);
asm
	jmp System.@InitializeArray
end;

procedure FinalizeArray(p: Pointer; typeInfo: Pointer; elemCount: Cardinal);
asm
	jmp System.@FinalizeArray
end;

procedure CopyArray(Dest, Source, TypeInfo: Pointer; Count: Integer);
asm
	jmp System.@CopyArray
end;

procedure AddRefArray(p: Pointer; typeInfo: Pointer; elemCount: Longint);
asm
	jmp System.@AddRefArray
end;	

procedure _LStrArrayClr(var StrArray; cnt: longint);
asm
	jmp System.@LStrArrayClr
end;

procedure _WStrArrayClr(var StrArray; Count: Integer);
asm
	jmp System.@WStrArrayClr
end;

{ Polymorphic dynamic array }

function DynArrayLength(A: Pointer): Integer;
asm
	jmp System.@DynArrayLength
end;

procedure DynArrayCopy(a: Pointer; typeInfo: Pointer; var Result: Pointer);
asm
	jmp System.@DynArrayCopy
end;

procedure DynArrayCopyRange(a: Pointer; typeInfo: Pointer; index, count : Integer; var Result: Pointer);
asm
	jmp System.@DynArrayCopyRange
end;

procedure DynArrayAsg(var A: Pointer; Source: Pointer; TypeInfo: PDynArrayTypeInfo);
asm
	jmp System.@DynArrayAsg
end;

procedure DynArrayAddRef(A: Pointer);
asm
	jmp System.@DynArrayAddRef
end;

procedure DynArrayUnique(var A: Pointer; TypeInfo: PDynArrayTypeInfo);
asm
	mov ecx, [eax]
	test ecx, ecx
	jz @Exit
	mov ecx, [ecx - 8]
	cmp ecx, 1
	jna @Exit
	mov ecx, eax
	mov eax, [eax]
	call System.@DynArrayCopy
@Exit:
end;

{ Variants }

procedure _VarClr(var v: TVarData);
asm
	jmp System._VarClr
end;

procedure _VarCopy(var Dest: TVarData; const Src: TVarData);
asm
	jmp System._VarCopy
end;

{ Interfaces }

function _IntfClear(var Dest: IInterface): Pointer;
asm
	jmp System.@IntfClear
end;

end.

