unit WinTips;

{*******************************************************}
{                                                       }
{       Delphi Run-time Library                         }
{       Windows API Tips Collection                     }
{                                                       }
{       1998..2006, 2008 YT                             }
{                                                       }
{*******************************************************}

{$WARN SYMBOL_PLATFORM OFF}

{x$DEFINE VERSION_ANIMATION}

interface

uses
	Windows, Messages, Multimon, Types, SysUtils, SysConst, {Imm,} 
	SysUtilsW, SysHacks;

{ **** kernel32.dll **** }

{ kernel32.dll - Console API }

const
	ATTACH_PARENT_PROCESS = DWORD(-1);

function GetConsoleWindow: THandle; stdcall; external kernel32;
function AttachConsole(ProcessId: DWORD): BOOL; stdcall; external kernel32;

{ kernel32.dll - Path API }

function GetLongPathNameA(ShortPath: PAnsiChar; LongPath: PAnsiChar; Buffer: DWORD): DWORD; stdcall; external kernel32;
function GetLongPathNameW(ShortPath: PWideChar; LongPath: PWideChar; Buffer: DWORD): DWORD; stdcall; external kernel32;

{ kernel32.dll - File System API }

type
	PREPARSE_MOUNTPOINT_DATA_BUFFER = ^REPARSE_MOUNTPOINT_DATA_BUFFER;
	REPARSE_MOUNTPOINT_DATA_BUFFER = record
		ReparseTag: DWORD;
		ReparseDataLength: WORD;
		Reserved: WORD;
		TargetNameOffset: WORD;
		TargetNameLength: WORD;
		PrintNameOffset: WORD;
		PrintNameLength: WORD;
		PathBuffer: array[0..0] of WideChar;
	end;

const
	INVALID_FILE_ATTRIBUTES = DWORD(-1);
	FILE_ATTRIBUTE_REPARSE_POINT = 1024;
	MAX_REPARSE_SIZE = 17000;
	FILE_FLAG_OPEN_REPARSE_POINT = $00200000;
	FSCTL_GET_REPARSE_POINT = $000900a8;
	IORPS_TAG_SYMBOLIC_LINK = $08000000;
	IORPS_TAG_MOUNT_POINT   = $A0000003;

function GetFileSizeEx(hFile: THandle; FileSize: PLargeInteger): BOOL; stdcall; external kernel32;

{ kernel32.dll - Thread API }

const
	 MWMO_INPUTAVAILABLE = $0004;

{ kernel32.dll - Security API }

type
	PTOKEN_USER = ^TOKEN_USER;
	TOKEN_USER = record
		User: SID_AND_ATTRIBUTES;
	end;

	ACL_SIZE_INFORMATION = record
		AceCount: DWORD;
		AclBytesInUse: DWORD;
		AclBytesFree: DWORD;
	end;

	PACE_HEADER = ^ACE_HEADER;
	ACE_HEADER = record
		AceType: BYTE;
		AceFlags: BYTE;
		AceSize: WORD;
	end;

	PACCESS_ALLOWED_ACE = ^ACCESS_ALLOWED_ACE;
	ACCESS_ALLOWED_ACE = record
		Header: ACE_HEADER;
		Mask: ACCESS_MASK;
		SidStart: DWORD;
	end;

const
	ACL_REVISION = 2;
	ACL_REVISION_DS = 4;

	ACL_REVISION1 = 1;
	ACL_REVISION2 = 2;
	ACL_REVISION3 = 3;
	ACL_REVISION4 = 4;
	MIN_ACL_REVISION = ACL_REVISION2;
	MAX_ACL_REVISION = ACL_REVISION4;

	ACCESS_ALLOWED_ACE_TYPE = 0;

	SECURITY_NT_AUTHORITY: SID_IDENTIFIER_AUTHORITY = (Value: (0,0,0,0,0,5)); // ntifs

	SECURITY_BUILTIN_DOMAIN_RID = $00000020;

	DOMAIN_ALIAS_RID_ADMINS = $00000220;

{ User Utilieis }

function GetUserName: WideString;

{ String Utilities }

function MapString(Flags: DWORD; const T: WideString): WideString;

{ Environment Variables }

function ExpandEnvironmentStrings(const S: WideString): WideString;

{ File Utilities }

function FileSystemExists(const Name: WideString): Boolean;
function IsRootDir(const Dir: WideString): Boolean;

function ConcatPath(const Path, FileName: WideString): WideString;

type
	TFileSystemLinkType = (ltNormal, ltJunction, ltHardLink);

function GetFileSystemLinkType(const FileName: WideString): TFileSystemLinkType; overload;
function GetFileSystemLinkType(const FileName: WideString; Attr: DWORD): TFileSystemLinkType; overload;

function ReplaceFile(const Source, Dest: WideString; New: Boolean = False): Boolean;

function GetVolumeLabel(const Drive: WideString): WideString;
function GetDiskFree(const Drive: WideString): Int64;

function SystemDrive: WideString;
function SystemDirectory: WideString;
function SystemPath: WideString;
function WindowsDirectory: WideString;
function WindowsPath: WideString;
function TemporaryDirectory: WideString;
function TemporaryPath: WideString;
function TemporaryFileName(const Directory: WideString): WideString; overload;
function TemporaryFileName: WideString; overload;

function UnixUTCToFileTime(UTC: DWORD): FILETIME;
function FileTimeToUnixUTC(const Time: FILETIME): DWORD;

{ Wildcard }

function MatchesWildCard(const FileName, WildCard: WideString): Boolean;

type
	TWildCardOperation = (woMatch, woAttribute, woAnd, woOr, woXor, woNot);
	TWildCardErrorKind = (weNone, weInvalidChar, weTerm, weOperator, weParenthesize);
	PWildCardNode = ^TWildCardNode;
	TWildCardNode = record
		WildCard: WideString;
		Attribute: Cardinal;
		Argument1: PWildCardNode;
		Argument2: PWildCardNode;
		Operation: TWildCardOperation;
	end;

type
	PWildCardExpression = ^TWildCardExpression;
	TWildCardExpression = object
	private
		FSource: WideString;
		FExpression: PWildCardNode;
	public
		constructor Initialize(const ASource: WideString; out Error: TWildCardErrorKind);
		destructor Finalize;
		function Evaluate(const FileName: WideString; Attribute: Cardinal): Boolean;
		property Source: WideString read FSource;
	end;

{ Handle Utilities }

function DuplicateHandle(Source: THandle): THandle;

{ File IO }

function GetFileSize(Handle: THandle): Int64;

function MapFile(Handle: THandle; Writable: Boolean;
	Offset: Int64; Size: Cardinal; out Mapping: THandle; out Address: Pointer): Boolean;
procedure Unmap(Mapping: THandle; Address: Pointer);

type
	TMapping = object
	private
		FileHandle: THandle;
	public
		Handle: THandle;
		Address: Pointer;
		Size: Cardinal;
		constructor Initialize(AFileHandle: THandle); overload;
		constructor Initialize(AFileName: WideString); overload;
		destructor Finalize;
	end;

{ Pipe Utilities }

function PipeAvailSize(Pipe: THandle): Cardinal;

{ Process Utilities }

function ExecuteCommandLine(
	const Command: WideString;
	const Directory: WideString = '';
	const ShowWindow: WORD = SW_NORMAL;
	const Wait: Boolean = False): Boolean; overload;

type
	TChildProcess = record
		Handle: THandle;
		StdIn: THandle;
		StdOut: THandle;
		StdErr: THandle;
	end;

function ExecuteCommandLine(
	const Command: WideString;
	const Directory: WideString;
	MergeOutAndErr: Boolean;
	out ChildProcess: TChildProcess): Boolean; overload;

procedure WaitFor(ChildProcess: TChildProcess); overload;

function ExecuteCommandLine(
	const Command: WideString;
	const Directory: WideString;
	const StdIn: string; out StdOut: string; StdErr: PString;
	ErrorLevel: PDWORD): Boolean; overload;

{ Version Information }

const
	vrCompanyName				= 'CompanyName';
	vrFileDescription		= 'FileDescription';
	vrFileVersion				= 'FileVersion';
	vrInternalName			= 'InternalName';
	vrLegalCopyright		= 'LegalCopyright';
	vrLegalTrademarks		= 'LegalTrademarks';
	vrOriginalFileName	= 'OriginalFilename';
	vrPrivateBuild			= 'PrivateBuild';
	vrProductName				= 'ProductName';
	vrProductVersion		= 'ProductVersion';
	vrSpecialBuild			= 'SpecialBuild';

type
	TStringLateInfo = record
		xLateLang: Smallint;
		xLateCode: Smallint;
	end;

	TVersionBuffer = record
		InfoBuffer: string;
		LateInfo: string;
	end;

function GetVersionBuffer(const FileName: TFileName; var Buffer: TVersionBuffer): Boolean;
function GetVersionString(const Buf: TVersionBuffer; const Ident: string): string;

{ Thread Utilities }

function CreateAutoEvent(Initial: Boolean): THandle;
function CreateManualEvent(Initial: Boolean): THandle;

type
	TWaitFlags = set of (
		wfWaitAll        =  0, {WMWO_WAITALL}
		wfAlertable      =  1, {MWMO_ALERTABLE}
		wfInputAvailable =  2, {MWMO_INPUTAVAILABLE}
		_wfHigh          = 31);

	TWaitResult = (
		wrSignaled     = WAIT_OBJECT_0,
		wrInput        = WAIT_ABANDONED_0 - 1,
		wrAbandoned    = WAIT_ABANDONED_0,
		wrIOCompletion = WAIT_IO_COMPLETION,
		wrTimeOut      = WAIT_TIMEOUT);

function WaitFor(const Handle: THandle;
	Milliseconds: DWORD = INFINITE): TWaitResult; overload;
function WaitFor(const Handles: array of THandle;
	var SignaledHandle: THandle;
	Milliseconds: DWORD = INFINITE;
	Flags: TWaitFlags = [];
	InputMask: DWORD = 0): TWaitResult; overload;

type
	TCriticalSection = object
	private
		Handle: TRTLCriticalSection;
	public
		constructor Initialize;
		destructor Finalize;
		function TryBeginWrite: Boolean;
		procedure BeginWrite;
		procedure EndWrite;
		procedure BeginRead;
		procedure EndRead;
		property OwningThread: THandle read Handle.OwningThread;
	end;

	TSemaphoreMultiReadExclusiveWriteSynchronizer = object
	private
		Handle: TRTLCriticalSection;
		ReadCount: Integer;
		FWriterEvent: THandle;
	public
		constructor Initialize;
		destructor Finalize;
		procedure BeginRead;
		procedure EndRead;
		function TryBeginWrite: Boolean;
		procedure BeginWrite;
		procedure EndWrite;
	end;

	CRendezvous = procedure(Context: Pointer) cdecl;

	TAtomicMultiReadExclusiveWriteSynchronizer = object
	public type
		TEntry = object
		private
			FCallerEntry: THandle;
			FCalleeEntry: THandle;
			FCalleeExit: THandle;
			FCallerState: Integer;
			FCalleeState: Integer;
			FBlock: CRendezvous;
			FContext: Pointer;
		public
			constructor Initialize; overload;
			constructor Initialize(Block: CRendezvous; Context: Pointer); overload;
			destructor Finalize;
		end;
	private
		FLocking: Integer;
		FWaitingWriters: Integer;
		FRevision: Cardinal;
		FThreadState: Integer; {TLS Index}
		FReaderEvent: THandle;
		FWriterEvent: THandle;
		function ThreadState: Integer;
		procedure PushThreadState(Value: Integer);
		procedure PopThreadState;
		procedure InternalBeginRead;
		procedure InternalEndRead;
		function InternalTryBeginWrite: Boolean;
		procedure InternalBeginWrite;
		procedure InternalEndWrite;
		function InternalElevating: Boolean;
	public
		constructor Initialize;
		destructor Finalize;
		procedure BeginRead;
		procedure EndRead;
		function TryBeginWrite: Boolean;
		function BeginWrite: Boolean;
		procedure EndWrite;
		procedure Async;
		procedure RestoreAsync;
		procedure Rendezvous(ThereThread: THandle; var Entry: TAtomicMultiReadExclusiveWriteSynchronizer.TEntry);
		procedure Accept(var Entry: TAtomicMultiReadExclusiveWriteSynchronizer.TEntry); overload;
		procedure Accept(var Entry: TAtomicMultiReadExclusiveWriteSynchronizer.TEntry;
			Block: CRendezvous; Context: Pointer); overload;
		property Revision: Cardinal read FRevision;
	end;

{ Memory Managers } 

type
	THeapMemoryManager = object
	private
		Heap: THandle;
	public
		constructor Initialize;
		destructor Finalize;
		function Allocate(ASize: Cardinal): Pointer;
		procedure Free(AUnit: Pointer);
	end;

	TFixedMemoryManager = object
	private
		Memory: Pointer;
		UnitSize: Cardinal;
		First: Pointer;
		Last: Pointer;
	public
		constructor Initialize(AUnitSize: Cardinal);
		destructor Finalize;
		function Allocate: Pointer;
		procedure Free(AUnit: Pointer);
	end;

{ **** user32.dll **** }

{ User32.dll - Format API }

function wsprintfW(Output, Format: PWideChar): Integer; cdecl; varargs; external User32;

{ User32.dll - Window Message }

const
	WM_NCXBUTTONDOWN   = $00AB;
	{$EXTERNALSYM WM_NCXBUTTONDOWN}
	WM_NCXBUTTONUP     = $00AC;
	{$EXTERNALSYM WM_NCXBUTTONUP}
	WM_NCXBUTTONDBLCLK = $00AD;
	{$EXTERNALSYM WM_NCXBUTTONDBLCLK}

type
	TWMNCXHitMessage = packed record
		Msg: Cardinal;
		HitTest: Word;
		Buttom: Word;
		XCursor: Smallint;
		YCursor: Smallint;
		Result: Longint;
	end;

{ Version Dialog }

function VersionDialog(ParentWnd: HWND; const Title: WideString): Boolean;

{ Screen Utilities }

function SystemWorkArea: TRect;

function GetFullScreenRect: TRect;
function GetActiveMonitorRect(Window: HWND = 0): TRect;

{ Window Utilities }

type
	CEachWindow = function(Wnd: HWND; Context: Pointer): Boolean cdecl;

procedure EnumerateWindows(Proc: CEachWindow; Context: Pointer);
procedure EnumerateChildWindows(Parent: HWND; Proc: CEachWindow; Context: Pointer);

type
	TWindowEnumerator = class(TEnumerator)
		function GetCurrent: HWND;
		property Current: HWND read GetCurrent;
	end;
	TWindowCollection = record
		function GetEnumerator: TWindowEnumerator;
		class function List: TWindowCollection; inline; static;
	end;
	TChildWindowCollection = record
		Parent: HWND;
		function GetEnumerator: TWindowEnumerator;
		class function List(AParent: HWND): TChildWindowCollection; inline; static;
	end;

{ Cursor Utilities }

procedure SetCursorWait;
procedure ResetCursor;

function DragDetect(const From, Current: TPoint): Boolean;

{ Keyboard Utilities }

function KeyNameText(VirtualKey: Byte): WideString;

{ **** gdi32.dll **** }

{ Bitmap Utilities } 

type
	PRGBQuadArray = ^TRGBQuadArray;
	TRGBQuadArray = array[0..32767] of TRGBQUAD;

	TBitmapInfoWithPalette = packed record
		bmiHeader: TBitmapInfoHeader;
		bmiColors: array[0..255] of TRGBQuad;
	end;

	TBitmapCoreInfoWithPalette = packed record
		bmciHeader: TBitmapCoreHeader;
		bmciColors: array[0..255] of TRGBTriple;
	end;

{ Text Utilities }

const
	StandardFontWidth = WideString('x');

function GetFontSize(Font: HFONT; const Text: WideString): TSize;
function GetTextSize(DC: HDC; const Text: WideString): TSize;

function TextOut(DC: HDC; X, Y: Integer; const Text: WideString): Boolean; inline;

function MinimizeText(const Text: WideString; DC: HDC; MaxLen: Integer): WideString;

function MinimizeName(const Filename: AnsiString; DC: HDC; MaxLen: Integer): AnsiString; overload;
function MinimizeName(const Filename: WideString; DC: HDC; MaxLen: Integer): WideString; overload;

{ **** imm32.dll **** }

{ imm32.dll - Command API }

type
	PRECONVERTSTRING = ^RECONVERTSTRING;
	RECONVERTSTRING = record
		dwSize: DWORD;
		dwVersion: DWORD;
		dwStrLen: DWORD;
		dwStrOffset: DWORD;
		dwCompStrLen: DWORD;
		dwCompStrOffset: DWORD;
		dwTargetStrLen: DWORD;
		dwTargetStrOffset: DWORD;
	end;

const
	IMR_COMPOSITIONWINDOW      = 1;
	IMR_CANDIDATEWINDOW        = 2;
	IMR_COMPOSITIONFONT        = 3;
	IMR_RECONVERTSTRING        = 4;
	IMR_CONFIRMRECONVERTSTRING = 5;
	IMR_QUERYCHARPOSITION      = 6;
	IMR_DOCUMENTFEED           = 7;

{ **** comctl32.dll **** } 

{ comctl32.dll - Window Message }

const
	UDM_SETPOS32 = WM_USER + 113;
	UDM_GETPOS32 = WM_USER + 114;

{ Flat Scrollbar Utilities }

type
	PScrollBarManager = ^TScrollBarManager;
	TScrollBarManager = record
		EnableScrollBar: function(hWnd: HWND; wSBflags, wArrows: UINT): BOOL; stdcall;
		ShowScrollBar: function(hWnd: HWND; wBar: Integer; bShow: BOOL): BOOL; stdcall;
		GetScrollInfo: function(hWnd: HWND; BarFlag: Integer; var ScrollInfo: TScrollInfo): BOOL; stdcall;
		SetScrollInfo: function(hWnd: HWND; BarFlag: Integer; const ScrollInfo: TScrollInfo; Redraw: BOOL): Integer; stdcall;
		GetScrollRange: function(hWnd: HWND; nBar: Integer; var lpMinPos, lpMaxPos: Integer): BOOL; stdcall;
		SetScrollRange: function(hWnd: HWND; nBar, nMinPos, nMaxPos: Integer; bRedraw: BOOL): BOOL; stdcall;
		GetScrollPos: function(hWnd: HWND; nBar: Integer): Integer; stdcall;
		SetScrollPos: function(hWnd: HWND; nBar, nPos: Integer; bRedraw: BOOL): Integer; stdcall;
		GetScrollProp: function(p1: HWND; propIndex: Integer; p3: PInteger): Bool; stdcall;
		SetScrollProp: function(p1: HWND; index: Integer; newValue: Integer; p4: Bool): Bool; stdcall;
		Initialize: function(Wnd: HWND): Bool; stdcall;
		Uninitialize: procedure(Wnd: HWND); stdcall;
	end;

function _GetScrollProp(p1: HWND; propIndex: Integer; p3: PInteger): Bool; stdcall;
function _SetScrollProp(p1: HWND; index: Integer; newValue: Integer; p4: Bool): Bool; stdcall;
function _InitializeSB(Wnd: HWND): Bool; stdcall;
procedure _UninitializeSB(Wnd: HWND); stdcall;

const
	_DefaultScrollBarManager: TScrollBarManager = (
		EnableScrollBar: Windows.EnableScrollBar;
		ShowScrollBar: Windows.ShowScrollBar;
		GetScrollInfo: Windows.GetScrollInfo;
		SetScrollInfo: Windows.SetScrollInfo;
		GetScrollRange: Windows.GetScrollRange;
		SetScrollRange: Windows.SetScrollRange;
		GetScrollPos: Windows.GetScrollPos;
		SetScrollPos: Windows.SetScrollPos;
		GetScrollProp: _GetScrollProp;
		SetScrollProp: _SetScrollProp;
		Initialize: _InitializeSB;
		Uninitialize: _UninitializeSB);

	DefaultScrollBarManager: PScrollBarManager = @_DefaultScrollBarManager;

function FlatScrollBarManager: PScrollBarManager;

implementation

{ User Utilieis }

function GetUserName: WideString;
var
	NameSize: DWORD;
begin
	NameSize := 256;
	SetLength(Result, NameSize);
	if not Windows.GetUserNameW(PWideChar(Result), NameSize) then RaiseLastOSError;
	SetLength(Result, NameSize - 1);
end;

{ String Utilities }

function MapString(Flags: DWORD; const T: WideString): WideString;
var
	Locale: DWORD;
begin
	Locale := GetUserDefaultLCID;
	SetLength(Result, LCMapStringW(Locale, Flags, Pointer(T), Length(T), nil, 0));
	LCMapStringW(Locale, Flags, Pointer(T), Length(T), Pointer(Result), Length(Result));
end;

{ Environment Variables }

function ExpandEnvironmentStrings(const S: WideString): WideString;
var
	Len: integer;
begin
	Result := '';
	Len := Windows.ExpandEnvironmentStringsW(PWideChar(S), nil, 0);
	if Len > 0 then
	begin
		SetLength(Result, Len - 1);
		Windows.ExpandEnvironmentStringsW(PWideChar(S), PWideChar(Result), Len);
	end;
end;

{ File Utilities }

function FileSystemExists(const Name: WideString): Boolean; overload;
var
	FD: WIN32_FIND_DATAW;
	H: THandle;
	V: array[0..MAX_PATH] of WideChar;
	Dummy: DWORD;
	E: UINT;
begin
	E := SetErrorMode(SEM_FAILCRITICALERRORS);
	try
		H := FindFirstFileW(PWideChar(Name), FD);
		if H = INVALID_HANDLE_VALUE then
			H := FindFirstFileW(PWideChar(IncludeTrailingPathDelimiter(Name) + '*.*'), FD);
		if H <> INVALID_HANDLE_VALUE then
		begin
			Windows.FindClose(H); {A/W}
			Result := True;
		end
		else
			Result := GetVolumeInformationW(PWideChar(Name), V, MAX_PATH, nil, Dummy, Dummy, nil, 0)
	finally
		SetErrorMode(E)
	end
end;

function IsRootDir(const Dir: WideString): Boolean;
begin
	if Dir = '' then
		Result := False
	else if Dir[1] = PathDelim then
		Result := ExtractFileDrive(Dir) = Dir {UNC}
	else if Length(Dir) = 3 then
		Result := (Dir[2] = DriveDelim) and (Dir[3] = PathDelim)
	else
		Result := False;
end;

function ConcatPath(const Path, FileName: WideString): WideString;
begin
	if (Length(FileName) > 2) and ((FileName[2] = ':') or (FileName[1] = '\')) then
		Result := FileName
	else
		Result := IncludeTrailingPathDelimiter(Path) + FileName
end;

function GetFileSystemLinkType(const FileName: WideString): TFileSystemLinkType; overload;
begin
	Result := GetFileSystemLinkType(FileName, GetFileAttributesW(PWideChar(FileName)))
end;

function GetFileSystemLinkType(const FileName: WideString; Attr: DWORD): TFileSystemLinkType; overload;
var
	Size: DWORD;
	ReparseBuffer: array[0..MAX_REPARSE_SIZE - 1] of Byte;
	Data: PREPARSE_MOUNTPOINT_DATA_BUFFER;
	F: THandle;
	R: Boolean;
begin
	if Attr = INVALID_FILE_ATTRIBUTES then
		RaiseLastOSError(ERROR_FILE_NOT_FOUND);

	if Attr and FILE_ATTRIBUTE_REPARSE_POINT = 0 then
		Result := ltNormal
	else
	begin
		Data := PREPARSE_MOUNTPOINT_DATA_BUFFER(@ReparseBuffer);
		F := CreateFileW(
			PWideChar(FileName),
			0,
			FILE_SHARE_READ or FILE_SHARE_WRITE,
			nil,
			OPEN_EXISTING,
			FILE_FLAG_BACKUP_SEMANTICS or FILE_FLAG_OPEN_REPARSE_POINT,
			0);
		if F = INVALID_HANDLE_VALUE then
			Result := ltNormal
		else
		begin
			R := DeviceIoControl(
				F,
				FSCTL_GET_REPARSE_POINT,
				nil,
				0,
				Data,
				Sizeof(ReparseBuffer),
				Size,
				nil);
			CloseHandle(F);
			if not R then
				Result := ltNormal
			else
			begin
				if Data^.ReparseTag = IORPS_TAG_SYMBOLIC_LINK then
					Result := ltJunction
				else if Data^.ReparseTag = IORPS_TAG_MOUNT_POINT then
					Result := ltHardLink
				else
					Result := ltNormal
			end
		end
	end
end;

function ReplaceFile(const Source, Dest: WideString; New: Boolean = False): Boolean;
var
	ReplaceFileAPI : function(
		ReplacedFileName: PWideChar;
		ReplacementFileName: PWideChar;
		BackupFileName: PWideChar;
		ReplaceFlags: DWORD;
		Exclude: Pointer;
		Reserved: Pointer): BOOL; stdcall;
begin
	ReplaceFileAPI := GetProcAddress(GetModuleHandle(kernel32), 'ReplaceFileW');
	if Assigned(ReplaceFileAPI) then
	begin
		Result := ReplaceFileAPI(PWideChar(Dest), PWideChar(Source), nil, 0, nil, nil);
		if Result or not New then Exit;
	end;
	Result := MoveFileExW(PWideChar(Source), PWideChar(Dest), MOVEFILE_REPLACE_EXISTING);
	if not Result then
	begin
		Sleep(250);
		Result := MoveFileExW(PWideChar(Source), PWideChar(Dest), MOVEFILE_REPLACE_EXISTING);
	end;
end;

function GetVolumeLabel(const Drive: WideString): WideString;
var
	Dummy: DWORD;
	Buffer: array[0..MAX_PATH] of WideChar;
begin
	if not GetVolumeInformationW(PWideChar(Drive), Buffer, MAX_PATH, nil, Dummy, Dummy, nil, 0) then
	begin
		RaiseLastOSError;
	end;
	Result := Buffer;
end;

function GetDiskFree(const Drive: WideString): Int64;
var
	Total: Int64;
begin
	if not GetDiskFreeSpaceExW(PWideChar(Drive), Result, Total, nil) then
		Result := -1
end;

function SystemDrive: WideString;
begin
	Result := ExtractFileDrive(SystemDirectory)
end;

function SystemDirectory: WideString;
var
	P: array[0..MAX_PATH] of WideChar;
begin
	SetString(Result, P, GetSystemDirectoryW(P, MAX_PATH));
end;

function SystemPath: WideString;
var
	P: array[0..MAX_PATH] of WideChar;
	Len: Integer;
begin
	Len := GetSystemDirectoryW(P, MAX_PATH);
	if P[Len - 1] <> '\' then begin P[Len] := '\'; Inc(Len) end;
	SetString(Result, P, Len);
end;

function WindowsDirectory: WideString;
var
	P: array[0..MAX_PATH] of WideChar;
begin
	SetString(Result, P, GetWindowsDirectoryW(P, MAX_PATH));
end;

function WindowsPath: WideString;
var
	P: array[0..MAX_PATH] of WideChar;
	Len: Integer;
begin
	Len := GetWindowsDirectoryW(P, MAX_PATH);
	if P[Len - 1] <> '\' then begin P[Len] := '\'; Inc(Len) end;
	SetString(Result, P, Len);
end;

function TemporaryDirectory: WideString;
var
	P: array[0..MAX_PATH] of WideChar;
	Len: Integer;
begin
	Len := GetTempPathW(MAX_PATH, P);
	if (Len > 3) and (P[Len - 1] = '\') then Dec(Len);
	SetString(Result, P, Len);
end;

function TemporaryPath: WideString;
var
	P: array[0..MAX_PATH] of WideChar;
	Len: Integer;
begin
	Len := GetTempPathW(MAX_PATH, P);
	if P[Len - 1] <> '\' then begin P[Len] := '\'; Inc(Len) end;
	SetString(Result, P, Len);
end;

function TemporaryFileName(const Directory: WideString): WideString;
var
	FileName: array[0..MAX_PATH] of WideChar;
begin
	GetTempFileNameW(PWideChar(Directory), '', 0, FileName);
	Result := FileName;
end;

function TemporaryFileName: WideString; overload;
begin
	Result := TemporaryFileName(TemporaryDirectory);
end;	

function UnixUTCToFileTime(UTC: DWORD): FILETIME;
const
	Diff = 11644473600;
begin
	Int64(Result) := (UTC + Diff) * Int64(10 * 1000 * 1000);
end;

function FileTimeToUnixUTC(const Time: FILETIME): DWORD;
const
	Diff = 11644473600;
begin
	Result := (Int64(Time) div (10 * 1000 * 1000)) - Diff;
end;

{ Wildcard }

function MatchesWildCard(const FileName, WildCard: WideString): Boolean;

	function Matches(PFileName, PWildCard: PWideChar): Boolean;
	begin
		case PWildCard^ of
		#0:
			Result := PFileName^ = #0;
		'*':
		begin
			Inc(PWildCard);
			while PFileName^ <> #0 do
			begin
				if Matches(PFileName, PWildCard) then
				begin
					Result := True;
					Exit;
				end;
				Inc(PFileName);
			end;
			Result := Matches(PFileName, PWildCard)
		end;
		'?':
		begin
			if PFileName^ = #0 then
				Result := False
			else
				Result := Matches(PFileName + 1, PWildCard + 1);
		end;
		'.':
		begin
			Result := ((PFileName^ = '.') and Matches(PFileName + 1, PWildCard + 1)) or
				((PFileName^ = #0) and (Pos('.', FileName) = 0) and Matches(PFileName, PWildCard + 1));
		end;
		else
			Result := (WideUpCase(PFileName^) = WideUpCase(PWildCard^)) and
				Matches(PFileName + 1, PWildCard + 1);
		end;
	end;

begin
	Result := Matches(PWideChar(FileName), PWideChar(WildCard))
end;

{ TWildCardExpression }

procedure NodeDestroy(Node: PWildCardNode);
begin
	if Node <> nil then
	begin
		NodeDestroy(Node^.Argument1);
		NodeDestroy(Node^.Argument2);
		Dispose(Node);
	end
end;

function TWildCardExpression.Evaluate(const FileName: WideString; Attribute: Cardinal): Boolean;
	function NodeEvaluate(Node: PWildCardNode; const FileName: WideString; Attribute: Cardinal): Boolean;
	begin
		case Node^.Operation of
		woMatch:
			Result := MatchesWildCard(FileName, Node^.WildCard);
		woAttribute:
			Result := Attribute and Node^.Attribute <> 0;
		woAnd:
			Result := NodeEvaluate(Node^.Argument1, FileName, Attribute) and
							NodeEvaluate(Node^.Argument2, FileName, Attribute);
		woOr:
			Result := NodeEvaluate(Node^.Argument1, FileName, Attribute) or
							NodeEvaluate(Node^.Argument2, FileName, Attribute);
		woXor:
			Result := NodeEvaluate(Node^.Argument1, FileName, Attribute) xor
							NodeEvaluate(Node^.Argument2, FileName, Attribute);
		else {woNot}
			Result := not NodeEvaluate(Node^.Argument1, FileName, Attribute);
		end;
	end;
begin
	Result := NodeEvaluate(FExpression, FileName, Attribute)
end;

destructor TWildCardExpression.Finalize;
begin
	NodeDestroy(FExpression);
	FExpression := nil;
end;

constructor TWildCardExpression.Initialize(const ASource: WideString; out Error: TWildCardErrorKind);

	var LastPlus1, P: PWideChar;

	type TWildCardToken = record
		Token: WideString;
		Kind: (wtNull, wtWord, wtString, wtSymbol);
	end;

	function Scan(var Source: PWideChar): TWildCardToken;
	var
		L: Integer;
		N: PWideChar;
		S: WideString;
	begin
		if Source = LastPlus1 then
		begin
			Result.Token := '';
			Result.Kind := wtNull;
		end
		else
		begin
			case Source^ of
			'"':
				begin
					{AnsiExtractQuotedStr"ĂꍇłɓĂ܂̂}
					S := Source + '#"';
					N := PWideChar(S);
					Result.Token := WideExtractQuotedStr(N, '"');
					L := N - PWideChar(S);
					if L = Length(S) then
					begin
						Error := weParenthesize;
						Exit;
					end;
					Result.Kind := wtString;
					Inc(Source, L);
				end;
			'A'..'Z', 'a'..'z':
				begin
					N := Source;
					repeat
						Inc(N);
						if N = LastPlus1 then Break;
						case N^ of
						'A'..'Z', 'a'..'z':
							;
						else
							Break;
						end;
					until False;
					SetString(Result.Token, Source, N - Source);
					Result.Kind := wtWord;
					Source := N;
				end;
			'(', ')', ';':
				begin
					Result.Token := Source^;
					Result.Kind := wtSymbol;
					Inc(Source);
				end;
			else
				Error := weInvalidChar;
			end;
		end;
		Source := WideStrTrimLeft(Source);
	end;

	function ParseOrXor(var Source: PWideChar): PWildCardNode; forward;

	function ParseMatch(var Source: PWideChar): PWildCardNode;
	var
		T: TWildCardToken;
		A: Cardinal;
	begin
		T := Scan(Source);
		if Error <> weNone then
		begin
			Result := nil;
		end
		else if T.Kind = wtNull then
		begin
			Error := weTerm;
			Result := nil;
		end
		else if T.Kind = wtString then
		begin
			New(Result);
			Result.Operation := woMatch;
			Result.Argument1 := nil;
			Result.Argument2 := nil;
			Result.WildCard := T.Token;
		end
		else if T.Kind = wtWord then
		begin
			if CompareText(T.Token, 'Hidden') = 0 then
				A := faHidden
			else if CompareText(T.Token, 'ReadOnly') = 0 then
				A := faReadOnly
			else if CompareText(T.Token, 'Archive') = 0 then
				A := faArchive
			else if CompareText(T.Token, 'System') = 0 then
				A := faSysFile
			else
			begin
				Error := weTerm;
				Result := nil;
				Exit;
			end;
			New(Result);
			Result.Operation := woAttribute;
			Result.Argument1 := nil;
			Result.Argument2 := nil;
			Result.Attribute := A;
		end
		else if T.Token = '(' then
		begin
			Result := ParseOrXor(Source);
			T := Scan(Source);
			if T.Token <> ')' then
			begin
				Error := weParenthesize;
				NodeDestroy(Result);
				Result := nil;
				Exit;
			end;
		end
		else
		begin
			Error := weTerm;
			Result := nil;
		end
	end;

	function ParseNot(var Source: PWideChar): PWildCardNode;
	var
		T: TWildCardToken;
		E: PWildCardNode;
		Next: PWideChar;
	begin
		Next := Source;
		T := Scan(Next);
		if (T.Kind = wtWord) and (CompareText(T.Token, 'NOT') = 0) then
		begin
			Source := Next;
			E := ParseMatch(Source);
			if E = nil then
			begin
				Result := nil;
				Exit;
			end;
			New(Result);
			Result.Operation := woNot;
			Result.Argument1 := E;
			Result.Argument2 := nil;
		end
		else
			Result := ParseMatch(Source);
	end;

	function ParseAnd(var Source: PWideChar): PWildCardNode;
	var
		T: TWildCardToken;
		E1, E2: PWildCardNode;
		Next: PWideChar;
	begin
		Result := ParseNot(Source);
		if Result = nil then Exit;
		repeat
			Next := Source;
			T := Scan(Next);
			if (T.Kind = wtWord) and (CompareText(T.Token, 'AND') = 0) then
			begin
				Source := Next;
				E1 := Result;
				E2 := ParseNot(Source);
				if E2 = nil then
				begin
					NodeDestroy(E1);
					Result := nil;
					Exit;
				end;
				New(Result);
				Result.Operation := woAnd;
				Result.Argument1 := E1;
				Result.Argument2 := E2;
			end
			else
				Break;
		until False;
	end;

	function ParseOrXor(var Source: PWideChar): PWildCardNode;
	var
		T: TWildCardToken;
		E1, E2: PWildCardNode;
		Next: PWideChar;
	begin
		Result := ParseAnd(Source);
		if Result = nil then Exit;
		repeat
			Next := Source;
			T := Scan(Next);
			if (T.Kind = wtWord) and (CompareText(T.Token, 'OR') = 0) or
							(T.Kind = wtSymbol) and (T.Token = ';') then
			begin
				Source := Next;
				E1 := Result;
				E2 := ParseAnd(Source);
				if E2 = nil then
				begin
					NodeDestroy(E1);
					Result := nil;
					Exit;
				end;
				New(Result);
				Result.Operation := woOr;
				Result.Argument1 := E1;
				Result.Argument2 := E2;
			end
			else if (T.Kind = wtWord) and (CompareText(T.Token, 'XOR') = 0) then
			begin
				Source := Next;
				E1 := Result;
				E2 := ParseAnd(Source);
				if E2 = nil then
				begin
					NodeDestroy(E1);
					Result := nil;
					Exit;
				end;
				New(Result);
				Result.Operation := woXOr;
				Result.Argument1 := E1;
				Result.Argument2 := E2;
			end
			else
				Break;
		until False;
	end;

begin
	Error := weNone;
	FSource := ASource;

	LastPlus1 := Pointer(FSource);
	Inc(LastPlus1, Length(FSource));

	P := WideStrTrimLeft(Pointer(FSource));
	FExpression := ParseOrXor(P);

	if P <> LastPlus1 then Error := weOperator;
	if Error <> weNone then Finalize;
end;

{ Handle Utilities }

function DuplicateHandle(Source: THandle): THandle;
var
	Cur: THandle;
begin
	Cur := GetCurrentProcess;
	Windows.DuplicateHandle(Cur, Source, Cur, @Result, 0, True, DUPLICATE_SAME_ACCESS);
end;

{ File IO }

function GetFileSize(Handle: THandle): Int64;
begin
	if not GetFileSizeEx(Handle, @Result) then
		Result := -1;
end;

function MapFile(Handle: THandle; Writable: Boolean;
	Offset: Int64; Size: Cardinal; out Mapping: THandle; out Address: Pointer): Boolean;
const
	Protects: array[Boolean] of DWORD = (PAGE_READONLY, PAGE_READWRITE);
	Accesses: array[Boolean] of DWORD = (FILE_MAP_READ, FILE_MAP_WRITE);
var
	FileSize: Int64;
begin
	if Size = 0 then
	begin
		Mapping := INVALID_HANDLE_VALUE;
		Address := nil;
		Result := False
	end
	else
	begin
		FileSize := Offset + Size;
		Mapping := CreateFileMappingW(Handle, nil, Protects[Writable],
			Int64Rec(FileSize).Hi, Int64Rec(FileSize).Lo, nil);
		if Mapping = INVALID_HANDLE_VALUE then
		begin
			Result := False;
		end
		else
		begin
			Address := MapViewOfFileEx(Mapping, Accesses[Writable],
				Int64Rec(Offset).Hi, Int64Rec(Offset).Lo, Size, nil);
			if Address = nil then
			begin
				CloseHandle(Mapping);
				Result := False
			end
			else
			begin
				Result := True;
			end;
		end
	end;
end;

procedure Unmap(Mapping: THandle; Address: Pointer);
begin
	UnmapViewOfFile(Address);
	CloseHandle(Mapping);
end;

{ TMapping }

destructor TMapping.Finalize;
begin
  Unmap(Handle, Address);
	if FileHandle <> 0 then FileClose(FileHandle);
end;

constructor TMapping.Initialize(AFileHandle: THandle);
begin
	FileHandle := 0;
	Size := GetFileSize(AFileHandle);
	if not MapFile(FileHandle, False, 0, Size, Handle, Address) then
    RaiseLastOSError
end;

constructor TMapping.Initialize(AFileName: WideString);
begin
	FileHandle := FileOpen(AFileName, fmOpenRead);
	Size := GetFileSize(FileHandle);
	if not MapFile(FileHandle, False, 0, Size, Handle, Address) then 
		RaiseLastOSError
end;

{ Pipe Utilities }

function PipeAvailSize(Pipe: THandle): Cardinal;
begin
	if not PeekNamedPipe(Pipe, nil, 0, nil, @Result, nil) then
    Result := 0;
end;

{ Process Utilities }

function ExecuteCommandLine(
	const Command: WideString;
	const Directory: WideString = '';
	const ShowWindow: WORD = SW_NORMAL;
	const Wait: Boolean = False): Boolean; overload;
var
	SInfo: STARTUPINFOW;
	PInfo: PROCESS_INFORMATION;
begin
	GetStartupInfoW(STARTUPINFOA(SInfo)); {--;}
	Result := CreateProcessW(nil, PWideChar(Command), nil, nil, False, 0, nil,
		Pointer(Directory), SInfo, PInfo);
	if Result then
	begin
		if Wait then WaitForSingleObject(PInfo.hProcess, INFINITE);
		CloseHandle(PInfo.hProcess);
		CloseHandle(PInfo.hThread);
	end
end;

function ExecuteCommandLine(
	const Command: WideString;
	const Directory: WideString;
	MergeOutAndErr: Boolean;
	out ChildProcess: TChildProcess): Boolean; overload;
const
	PipeSecurity: TSecurityAttributes = (
		nLength: SizeOf(TSecurityAttributes);
		lpSecurityDescriptor: nil;
		bInheritHandle: True);
var
	StdInR, StdInW: THandle;
	StdOutR, StdOutW: THandle;
	StdErrR, StdErrW: THandle;
	SInfo: STARTUPINFOW;
	PInfo: PROCESS_INFORMATION;
begin
	CreatePipe(StdInR, StdInW, @PipeSecurity, 0);
	CreatePipe(StdOutR, StdOutW, @PipeSecurity, 0);
	StdErrR := 0;
	StdErrW := 0;
	if not MergeOutAndErr then CreatePipe(StdErrR, StdErrW, @PipeSecurity, 0);

	Windows.DuplicateHandle(GetCurrentProcess, StdInW, GetCurrentProcess,
		@StdInW, 0, False, DUPLICATE_SAME_ACCESS or DUPLICATE_CLOSE_SOURCE);

	GetStartupInfoW(STARTUPINFOA(SInfo));
	SInfo.dwFlags := STARTF_USESTDHANDLES or STARTF_FORCEOFFFEEDBACK or STARTF_USESHOWWINDOW;
	SInfo.hStdInput := StdInR;
	SInfo.hStdOutput := StdOutW;
	if not MergeOutAndErr then SInfo.hStdError := StdErrW else SInfo.hStdError := StdOutW;
	SInfo.wShowWindow := SW_MINIMIZE;

	Result := CreateProcessW(nil, PWideChar(Command), @PipeSecurity, nil, True,
		0, nil, Pointer(Directory), SInfo, PInfo);
	if Result then
	begin
		WaitForInputIdle(PInfo.hProcess, 1000);
		ChildProcess.Handle := PInfo.hProcess;
		ChildProcess.StdIn := StdInW;
		ChildProcess.StdOut := StdOutR;
		ChildProcess.StdErr := StdErrR;
	end
	else
	begin
		CloseHandle(PInfo.hProcess);
		CloseHandle(StdInW);
		CloseHandle(StdOutR);
		CloseHandle(StdErrR);
	end;
	CloseHandle(PInfo.hThread);
	CloseHandle(StdInR);
	CloseHandle(StdOutW);
	CloseHandle(StdErrW);
end;

procedure WaitFor(ChildProcess: TChildProcess); overload;
begin
	WaitFor(ChildProcess.Handle);
  CloseHandle(ChildProcess.Handle);
	CloseHandle(ChildProcess.StdIn);
	CloseHandle(ChildProcess.StdOut);
  CloseHandle(ChildProcess.StdErr); {0邪˂؂}
end;

function ExecuteCommandLine(
	const Command: WideString;
	const Directory: WideString;
	const StdIn: string; out StdOut: string; StdErr: PString;
	ErrorLevel: PDWORD): Boolean; overload;
var
	ChildProcess: TChildProcess;
	Buffer: array[0..1024] of WideChar;
	Written, Read: DWORD;
	From: Integer;
	Terminated: Boolean;
begin
	Result := ExecuteCommandLine(Command, Directory, StdErr = nil, ChildProcess);
	if not Result then Exit;
	try
		StdOut := '';
		if StdErr <> nil then StdErr^ := '';
		WriteFile(ChildProcess.StdIn, Pointer(StdIn)^, Length(StdIn), Written, nil);
		repeat
			Terminated := WaitFor(ChildProcess.Handle, 0) = wrSignaled;

			repeat
				if not ReadFile(ChildProcess.StdOut, Buffer, SizeOf(Buffer), Read, nil)
					and (GetLastError = ERROR_BROKEN_PIPE) then
				begin
					Break;
				end;
				From := Length(StdOut);
				SetLength(StdOut, From + Integer(Read));
				Move(Buffer, (PChar(StdOut) + From)^, Read);
				Dec(Written, Read);
			until PipeAvailSize(ChildProcess.StdOut) = 0;

			repeat
				if not ReadFile(ChildProcess.StdErr, Buffer, SizeOf(Buffer), Read, nil)
					and (GetLastError = ERROR_BROKEN_PIPE) then
				begin
					Break;
				end;
				if StdErr <> nil then
				begin
					From := Length(StdErr^);
					SetLength(StdErr^, From + Integer(Read));
					Move(Buffer, (PChar(StdErr^) + From)^, Read);
				end
			until PipeAvailSize(ChildProcess.StdErr) = 0;

		until Terminated;
	finally
		if ErrorLevel <> nil then
			GetExitCodeProcess(ChildProcess.Handle, ErrorLevel^);

		WaitFor(ChildProcess);
	end;
end;

{ Version Information }

function GetVersionBuffer(const FileName: TFileName; var Buffer: TVersionBuffer): Boolean;
const
	VerRes = '\VarFileInfo\Translation\';
var
	PFileName: PChar;
	Size, InfoHandle, Len: DWORD;
	LateLang, LateCode: Integer;
	Addr: Pointer;
begin
	PFileName := PChar(FileName);

	Size:= GetFileVersionInfoSize(PFileName, InfoHandle);
	Result := Size > 0;
	if Result then
	begin
		SetLength(Buffer.InfoBuffer, Size);
		if GetFileVersionInfo(PFileName, InfoHandle, Size, Pointer(Buffer.InfoBuffer)) then
		begin
			if VerQueryValue(Pointer(Buffer.InfoBuffer), VerRes, Addr, Len) then
			begin
				LateLang:= TStringLateInfo(Addr^).xLateLang;
				LateCode:= TStringLateInfo(Addr^).xLateCode;
			end
			else
			begin
				LateLang:= $0409;
				LateCode:= $0000;
			end;
			Buffer.LateInfo:= IntToHex(LateLang, 4) +IntToHex(LateCode, 4);
		end;
	end;
end;

function GetVersionString(const Buf: TVersionBuffer; const Ident: string): string;
var
	ResultAddr: Pointer;
	Len: LongWord;
begin
	if VerQueryValue(Pointer(Buf.InfoBuffer),
		Pointer('\StringFileInfo\' + Buf.LateInfo + '\' + Ident),
		ResultAddr, Len) then
	begin
		SetString(Result, PChar(ResultAddr), Len - 1);
	end
end;

{ Thread Utilities }

function CreateAutoEvent(Initial: Boolean): THandle;
begin
	Result := CreateEventW(nil, False, Initial, nil)
end;

function CreateManualEvent(Initial: Boolean): THandle;
begin
	Result := CreateEventW(nil, True, Initial, nil)
end;

function WaitFor(const Handle: THandle;
	Milliseconds: DWORD = INFINITE): TWaitResult; overload;
var
	R: DWORD;
begin
	R := WaitForSingleObject(Handle, Milliseconds);
	if R = WAIT_FAILED then RaiseLastOSError;
	Result := TWaitResult(R);
end;

function WaitFor(const Handles: array of THandle;
	var SignaledHandle: THandle;
	Milliseconds: DWORD = INFINITE;
	Flags: TWaitFlags = [];
	InputMask: DWORD = 0): TWaitResult;
var
	R: DWORD;
begin
	if InputMask = 0 then
  	{InputMask = 0łMsgWaitForSingleObjectsEx̓CxgɂԂĂ鎞}
		R := WaitForMultipleObjectsEx(
			Length(Handles), PWOHandleArray(@Handles[Low(Handles)]),
      wfWaitAll in Flags, Milliseconds, wfAlertable in Flags)
	else
		R := MsgWaitForMultipleObjectsEx(
			Length(Handles), (@Handles[Low(Handles)])^,
			Milliseconds, InputMask, DWORD(Flags));

	if R = WAIT_FAILED then RaiseLastOSError;

	if R = WAIT_OBJECT_0 + DWORD(Length(Handles)) then
		Result := wrInput
	else if R < 128 then
	begin
		SignaledHandle := Handles[R];
		Result := wrSignaled;
	end
	else if R < 256 then
	begin
		SignaledHandle := Handles[R - WAIT_ABANDONED_0];
		Result := wrAbandoned;
	end
	else
		Result := TWaitResult(R);
end;

{ TCriticalSection }

constructor TCriticalSection.Initialize;
begin
	InitializeCriticalSection(Handle);
end;

destructor TCriticalSection.Finalize;
begin
	DeleteCriticalSection(Handle);
end;

procedure TCriticalSection.BeginRead;
begin
	EnterCriticalSection(Handle);
end;

procedure TCriticalSection.BeginWrite;
begin
	EnterCriticalSection(Handle);
end;

procedure TCriticalSection.EndRead;
begin
	LeaveCriticalSection(Handle);
end;

procedure TCriticalSection.EndWrite;
begin
	LeaveCriticalSection(Handle);
end;

function TCriticalSection.TryBeginWrite: Boolean;
begin
	Result := TryEnterCriticalSection(Handle);
end;

{ TSemaphoreMultiReadExclusiveWriteSynchronizer }

procedure TSemaphoreMultiReadExclusiveWriteSynchronizer.BeginRead;
begin
	EnterCriticalSection(Handle);
	Inc(ReadCount);
	LeaveCriticalSection(Handle);
end;

procedure TSemaphoreMultiReadExclusiveWriteSynchronizer.BeginWrite;
begin
	EnterCriticalSection(Handle);
	while ReadCount > 0 do
	begin
		LeaveCriticalSection(Handle);
		WaitFor(FWriterEvent);
		EnterCriticalSection(Handle);
	end;
end;

procedure TSemaphoreMultiReadExclusiveWriteSynchronizer.EndRead;
begin
	EnterCriticalSection(Handle);
	Dec(ReadCount);
	LeaveCriticalSection(Handle);
	SetEvent(FWriterEvent);
end;

procedure TSemaphoreMultiReadExclusiveWriteSynchronizer.EndWrite;
begin
	LeaveCriticalSection(Handle);
	SetEvent(FWriterEvent);
end;

destructor TSemaphoreMultiReadExclusiveWriteSynchronizer.Finalize;
begin
	CloseHandle(FWriterEvent);
	DeleteCriticalSection(Handle);
end;

constructor TSemaphoreMultiReadExclusiveWriteSynchronizer.Initialize;
begin
	FWriterEvent := CreateAutoEvent(False);
	InitializeCriticalSection(Handle);
end;

function TSemaphoreMultiReadExclusiveWriteSynchronizer.TryBeginWrite: Boolean;
begin
	Result := TryEnterCriticalSection(Handle);
	if Result then
	begin
		Result := ReadCount = 0;
		if not Result then LeaveCriticalSection(Handle);
	end;
end;

{ TAtomicMultiReadExclusiveWriteSynchronizer }

procedure TAtomicMultiReadExclusiveWriteSynchronizer.Accept(
	var Entry: TAtomicMultiReadExclusiveWriteSynchronizer.TEntry;
	Block: CRendezvous; Context: Pointer);
var
	Elevating: Boolean;
begin
	//OutputDebugString(PChar('ACCEPT ' + IntToStr(GetCurrentThreadId)));
	Entry.FCalleeState := ThreadState;
	SetEvent(Entry.FCalleeEntry);
	//OutputDebugString(PChar('ACCEPT1 ' + IntToStr(GetCurrentThreadId)));
	WaitFor(Entry.FCallerEntry);
	//OutputDebugString(PChar('ACCEPT2 ' + IntToStr(GetCurrentThreadId)));
	Elevating := (Entry.FCalleeState = 0) and (Entry.FCallerState <> 0);
	if Elevating then PushThreadState(Entry.FCallerState);
	if Assigned(Block) then Block(Context);
	if Elevating then PopThreadState;
	SetEvent(Entry.FCalleeExit);
	//OutputDebugString(PChar('END ACCEPT ' + IntToStr(GetCurrentThreadId)));
end;

procedure TAtomicMultiReadExclusiveWriteSynchronizer.Accept(
	var Entry: TAtomicMultiReadExclusiveWriteSynchronizer.TEntry);
begin
	Accept(Entry, Entry.FBlock, Entry.FContext);
end;

procedure TAtomicMultiReadExclusiveWriteSynchronizer.Async;
var
	State: Integer;
begin
	State := ThreadState;
	if State > 0 then
		InternalEndRead
	else if State < 0 then
		InternalEndWrite;

	PushThreadState(0);
end;

procedure TAtomicMultiReadExclusiveWriteSynchronizer.BeginRead;
var
	State: Integer;
begin
	State := ThreadState;
	if State <> 0 then
		PushThreadState(State + 1)
	else
	begin
		InternalBeginRead;
		PushThreadState(1);
	end;
end;

function TAtomicMultiReadExclusiveWriteSynchronizer.BeginWrite: Boolean;
var
	State: Integer;
	OldRevision: Cardinal;
begin
  OldRevision := Revision;
	State := ThreadState;
	if State < 0 then
		PushThreadState(State + 1)
	else
	begin
		if State > 0 then InternalEndRead;
		InternalBeginWrite;
		PushThreadState(Low(Integer));
	end;
	Result := OldRevision = Revision;
end;

procedure TAtomicMultiReadExclusiveWriteSynchronizer.EndRead;
var
	State: Integer;
begin
	State := ThreadState;
	if State = 1 then InternalEndRead;
	PopThreadState;
end;

procedure TAtomicMultiReadExclusiveWriteSynchronizer.EndWrite;
begin
	if ThreadState = Low(Integer) then InternalEndWrite;
	PopThreadState;
	if ThreadState > 0 then InternalBeginRead;
end;

destructor TAtomicMultiReadExclusiveWriteSynchronizer.Finalize;
begin
	TlsFree(FThreadState);
end;

constructor TAtomicMultiReadExclusiveWriteSynchronizer.Initialize;
begin
	FLocking := 0;
	FRevision := 0;
	FWaitingWriters := 0;
	FThreadState := TlsAlloc;
	FReaderEvent := CreateManualEvent(True);
	FWriterEvent := CreateAutoEvent(False);
end;

procedure TAtomicMultiReadExclusiveWriteSynchronizer.InternalBeginRead;
begin
	while InterlockedIncrement(FLocking) <= 0 do
	begin
		InterlockedDecrement(FLocking);
		WaitFor(FReaderEvent);
	end;
end;

procedure TAtomicMultiReadExclusiveWriteSynchronizer.InternalBeginWrite;
var
	Locking: Integer;
begin
  InterlockedIncrement(FWaitingWriters);
	repeat
		ResetEvent(FReaderEvent);
		Locking := InterlockedCompareExchange(FLocking, Low(Integer), 0);
		if Locking = 0 then Break;
		if Locking > 0 then SetEvent(FReaderEvent);
		WaitFor(FWriterEvent);
	until False;
  InterlockedDecrement(FWaitingWriters);
	ResetEvent(FReaderEvent);
	ResetEvent(FWriterEvent);
end;

function TAtomicMultiReadExclusiveWriteSynchronizer.InternalElevating: Boolean;
var
	Locking: Integer;
begin
	Locking := InterlockedCompareExchange(FLocking, Low(Integer), 1);
	Result := Locking = 1;
	if Result then
	begin
		ResetEvent(FReaderEvent);
		ResetEvent(FWriterEvent);
	end;
end;

procedure TAtomicMultiReadExclusiveWriteSynchronizer.InternalEndRead;
begin
	InterlockedDecrement(FLocking);
	if InterlockedExchangeAdd(FWaitingWriters, 0) > 0 then
		SetEvent(FWriterEvent);
end;

procedure TAtomicMultiReadExclusiveWriteSynchronizer.InternalEndWrite;
begin
	Inc(FRevision);
	InterlockedExchange(FLocking, 0);
	if InterlockedExchangeAdd(FWaitingWriters, 0) > 0 then
		SetEvent(FWriterEvent)
	else
		SetEvent(FReaderEvent);
end;

function TAtomicMultiReadExclusiveWriteSynchronizer.InternalTryBeginWrite: Boolean;
var
	Locking: Integer;
begin
	Locking := InterlockedCompareExchange(FLocking, Low(Integer), 0);
	Result := Locking = 0;
	if Result then
	begin
		ResetEvent(FReaderEvent);
		ResetEvent(FWriterEvent);
	end;
end;

procedure TAtomicMultiReadExclusiveWriteSynchronizer.PopThreadState;
var
	X: PInteger;
begin
	X := TlsGetValue(FThreadState);
	Assert(X <> nil);
	Assert(X^ > 0);
	Dec(X^);
	if X^ = 0 then
	begin
		FreeMem(X);
		TlsSetValue(FThreadState, nil);
  end;	
end;

procedure TAtomicMultiReadExclusiveWriteSynchronizer.PushThreadState(
	Value: Integer);
var
	X: PInteger;
	Len: Integer;
begin
	X := TlsGetValue(FThreadState);
	Len := 0;
	if X <> nil then Len := X^;
	Inc(Len);
	ReallocMem(X, (Len + 1) * SizeOf(Integer));
	TlsSetValue(FThreadState, X);
	X^ := Len;
	Inc(X, Len);
	X^ := Value;
end;

procedure TAtomicMultiReadExclusiveWriteSynchronizer.Rendezvous(
	ThereThread: THandle;
	var Entry: TAtomicMultiReadExclusiveWriteSynchronizer.TEntry);
var
	ReadRead: Boolean;
	Handles: array[0..1] of THandle;
	Signed: THandle;
begin
	Handles[1] := ThereThread;
	//OutputDebugString(PChar('RENDEZVOUS ' + IntToStr(GetCurrentThreadId)));
	Entry.FCallerState := ThreadState;
	Handles[0] := Entry.FCalleeEntry;
	WaitFor(Handles, Signed);
	//OutputDebugString(PChar('RENDEZVOUS1 ' + IntToStr(GetCurrentThreadId)));
	SetEvent(Entry.FCallerEntry);
	//OutputDebugString(PChar('RENDEZVOUS2 ' + IntToStr(GetCurrentThreadId)));
	ReadRead := (Entry.FCallerState > 0) and (Entry.FCalleeState > 0);
	if ReadRead then InternalEndRead;
	Handles[0] := Entry.FCalleeExit;
	WaitFor(Handles, Signed);
	//OutputDebugString(PChar('RENDEZVOUS3 ' + IntToStr(GetCurrentThreadId)));
	if ReadRead then InternalBeginRead;
	//OutputDebugString(PChar('END RENDEZVOUS ' + IntToStr(GetCurrentThreadId)));
end;

procedure TAtomicMultiReadExclusiveWriteSynchronizer.RestoreAsync;
var
	State: Integer;
begin
	PopThreadState;
	State := ThreadState;
	
	if State > 0 then
		InternalBeginRead
	else if State < 0 then
		InternalBeginWrite;
end;

function TAtomicMultiReadExclusiveWriteSynchronizer.ThreadState: Integer;
var
	X: PInteger;
	Index: Integer;
begin
	X := TlsGetValue(FThreadState);
	if X = nil then
		Result := 0
	else
	begin
		Index := X^;
		Inc(X, Index);
		Result := X^;
	end;
end;

function TAtomicMultiReadExclusiveWriteSynchronizer.TryBeginWrite: Boolean;
var
	State, NewState: Integer;
begin
	State := ThreadState;

	if State < 0 then
	begin
		Result := True;
		NewState := State + 1;
	end
	else if State > 0 then
	begin
		Result := InternalElevating;
		NewState := Low(Integer);
	end
	else
	begin
		Result := InternalTryBeginWrite;
		NewState := Low(Integer);
	end;

	if Result then PushThreadState(NewState);
end;

{ TAtomicMultiReadExclusiveWriteSynchronizer.TEntry }

destructor TAtomicMultiReadExclusiveWriteSynchronizer.TEntry.Finalize;
begin
	CloseHandle(FCallerEntry);
	CloseHandle(FCalleeEntry);
	CloseHandle(FCalleeExit);
end;

constructor TAtomicMultiReadExclusiveWriteSynchronizer.TEntry.Initialize;
begin
	Initialize(nil, nil);
end;

constructor TAtomicMultiReadExclusiveWriteSynchronizer.TEntry.Initialize(
	Block: CRendezvous; Context: Pointer);
begin
	FBlock := Block;
	FContext := Context;
	FCallerState := 0;
	FCalleeState := 0;
	FCallerEntry := CreateManualEvent(False);
	FCalleeEntry := CreateManualEvent(False);
	FCalleeExit := CreateManualEvent(False);
end;

{ Memory Managers } 

{ THeapMemoryManager }

function THeapMemoryManager.Allocate(ASize: Cardinal): Pointer;
begin
	Result := HeapAlloc(Heap, HEAP_NO_SERIALIZE, ASize);
	if Result = nil then
		OutOfMemoryError;
end;

destructor THeapMemoryManager.Finalize;
begin
	if Heap <> 0 then
		HeapDestroy(Heap)
end;

procedure THeapMemoryManager.Free(AUnit: Pointer);
begin
	if not HeapFree(Heap, HEAP_NO_SERIALIZE, AUnit) then
		RaiseLastOSError;
end;

constructor THeapMemoryManager.Initialize;
begin
	Heap := HeapCreate(HEAP_NO_SERIALIZE, 0, 0);
	if Heap = 0 then
		OutOfMemoryError;
end;

{ TFixedMemoryManager }

function TFixedMemoryManager.Allocate: Pointer;
var
	NewCommitSize: Integer;
	Next, I: Pointer;
begin
	if First = Last then
	begin
		NewCommitSize := UnitSize * 1024;

		if VirtualAlloc(Last, NewCommitSize, MEM_COMMIT, PAGE_READWRITE) = nil then
			OutOfMemoryError;

		Result := First;

		Inc(Cardinal(Last), NewCommitSize);
		Inc(Cardinal(First), UnitSize);

		I := First;
		while Cardinal(I) < Cardinal(Last) do
		begin
			Next := Pointer(Cardinal(I) + UnitSize);
			PPointer(I)^ := Next;
			I := Next;
		end;
	end
	else
	begin
		Result := First;
		First := PPointer(Result)^;
	end;
end;

destructor TFixedMemoryManager.Finalize;
var
	Commited: Cardinal;
begin
	Commited := Cardinal(Last) - Cardinal(Memory);
	if Commited <> 0 then
		VirtualFree(Memory, Commited, MEM_DECOMMIT);
	if Memory <> nil then
	begin
{$IFOPT C+}
		if not VirtualFree(Memory, 0, MEM_RELEASE) then RaiseLastOSError;
{$ELSE}
		VirtualFree(Memory, 0, MEM_RELEASE);
{$ENDIF}
	end;
end;

procedure TFixedMemoryManager.Free(AUnit: Pointer);
begin
	PPointer(AUnit)^ := First;
	First := AUnit;
end;

constructor TFixedMemoryManager.Initialize(AUnitSize: Cardinal);
const
	MaxLogicalLimitation = 1024 * 1024 * 1024; {1GB!}
var
	LogicalLimitation: Cardinal;
begin
	UnitSize := AUnitSize;
	if AUnitSize < SizeOf(Pointer) then
		UnitSize := SizeOf(Pointer);

	First := nil;
	Last := nil;

	LogicalLimitation := MaxLogicalLimitation;
	repeat
		Memory := VirtualAlloc(nil, LogicalLimitation, MEM_RESERVE, PAGE_READWRITE);
		if Memory <> nil then Break;
		LogicalLimitation := LogicalLimitation div 2;
		if LogicalLimitation < AUnitSize then
			OutOfMemoryError;
	until False;

	First := Memory;
	Last := Memory;
end;

{ Version Dialog }

type
{$IFDEF VERSION_ANIMATION}
	TAnimationMode = (amNone, amSin, amCircle, amCircleRev, amTeleport);
{$ENDIF}

	PVerWndField = ^TVerWndField;
	TVerWndField = record
		V_Company: string;
		V_Product: string;
		V_Version: string;
		V_Copyright: string;
{$IFDEF VERSION_DESCRIPTION}
		V_Comments: string;
{$ENDIF}
		WinVer: string;
		Memory: string;
		OkButton: HWND;
		Icon: HICON;
{$IFDEF VERSION_ANIMATION}
		AnimationMode: TAnimationMode;
{$ENDIF}
		Result: Boolean;
	end;

const
	DlgWidth  = 410;
	DlgHeight = 210;

	DlgExStyle = WS_EX_WINDOWEDGE or WS_EX_DLGMODALFRAME or WS_EX_TOPMOST;
	DlgStyle = WS_CAPTION or WS_SYSMENU or WS_POPUP or WS_CLIPSIBLINGS or WS_DLGFRAME;

	LinePos = 135;
	PictureBorder: TRect = (Left:5; Top:5; Right:105; Bottom:205);
	LineBorder: TRect = (Left:110; Top:LinePos - 2;
					Right:DlgWidth - 5; Bottom:LinePos);

{$IFDEF VERSION_ANIMATION}
	CenterX = (5 + 105) div 2;
	CenterY = (5 + 205) div 2;
	IDT_ANIMATE = 10;
{$ENDIF}

function VerWndProc(Wnd: HWND; Msg: UINT; wP: wParam; lP: lParam): LRESULT;
				stdcall;
var
	PS: PAINTSTRUCT;
	DC: HDC;
	Pen: HPEN;
	R: TRect;
	Field: PVerWndField;
	IconX, IconY: Integer;
{$IFDEF VERSION_ANIMATION}
	OrgRandSeed: Integer;
{$ENDIF}
	Y: Integer;
	H: Cardinal;
	BkColor1, BkColor2: COLORREF;
begin
	Field:= Pointer(GetWindowLongW(Wnd, GWL_USERDATA));

	case Msg of
	WM_PAINT:
	begin
		if wP = 0 then DC:= BeginPaint(Wnd, PS) else DC:= wP;
		SaveDC(DC);

		R:= PictureBorder;
		DrawEdge(DC, R, BDR_SUNKENOUTER, BF_ADJUST or BF_RECT);

		BkColor1 := GetSysColor(COLOR_ACTIVECAPTION);
		BkColor2 := GetSysColor(COLOR_GRADIENTACTIVECAPTION{COLOR_INFOBK});

		H := R.Bottom - R.Top;
		for Y := H - 1 downto 0 do
		begin
			Pen := CreatePen(PS_SOLID, 1,
							((BkColor1 and $0000ff) * (H - Cardinal(Y)) div H +
							(BkColor2 and $0000ff) * Cardinal(Y) div H) +
							((BkColor1 and $00ff00 shr 8) * (H - Cardinal(Y)) div H +
							(BkColor2 and $00ff00 shr 8) * Cardinal(Y) div H) shl 8 +
							((BkColor1 and $ff0000 shr 16) * (H - Cardinal(Y)) div H +
							(BkColor2 and $ff0000 shr 16) * Cardinal(Y) div H) shl 16);
			SaveDC(DC);
			SelectObject(DC, Pen);
			MoveToEx(DC, R.Left, Y + R.Top, nil);
			LineTo(DC, R.Right, Y + R.Top);
			RestoreDC(DC, -1);
			DeleteObject(Pen);
		end;

{$IFDEF VERSION_ANIMATION}
		case Field^.AnimationMode of
		amSin:
		begin
			IconX:= CenterX - 16;
			IconY:= CenterY - 16 + Trunc(Sin(GetTickCount / 1000) * 50);
		end;
		amCircle:
		begin
			IconX:= CenterX - 16 + Trunc(Cos(GetTickCount / 500) * 20);
			IconY:= CenterY - 16 + Trunc(Sin(GetTickCount / 500) * 20);
		end;
		amCircleRev:
		begin
			IconX:= CenterX - 16 + Trunc(Cos(GetTickCount / 500) * 20);
			IconY:= CenterY - 16 - Trunc(Sin(GetTickCount / 500) * 20);
		end;
		amTeleport:
		begin
			OrgRandSeed:= RandSeed;
			Randomize;
			IconX:= Random(R.Right - R.Left - 32) + R.Left;
			IconY:= Random(R.Bottom - R.Top - 32) + R.Top;
			RandSeed:= OrgRandSeed;
		end;
		else
			IconX:= R.Left + 10;
			IconY:= R.Top  + 10;
		end;
{$ELSE}
		IconX:= R.Left + 10;
		IconY:= R.Top  + 10;
{$ENDIF}

		DrawIcon(DC, IconX, IconY, Field^.Icon);

		SetBkColor(DC, GetSysColor(COLOR_3DFACE));
		TextOut(DC, 110,   5, Field.V_Company);
		TextOut(DC, 110,  25, Field.V_Product);
		TextOut(DC, 110,  45, Field.V_Version);
		TextOut(DC, 110,  65, Field.V_Copyright);
{$IFDEF VERSION_DESCRIPTION}
		TextOut(DC, 110,  85, Field.V_Comments);
{$ENDIF}
		SelectObject(DC, GetStockObject(DEFAULT_GUI_FONT));
{		SetTextColor(DC, GetSysColor(COLOR_GRAYTEXT));
		SetTextAlign(DC, TA_BOTTOM);
		TextOut(DC, 110, LinePos - 4, ParamStr(1));
		SetTextColor(DC, GetSysColor(COLOR_WINDOWTEXT));
		SetTextAlign(DC, TA_TOP);
}		R:= LineBorder;
		DrawEdge(DC, R, BDR_SUNKENOUTER, BF_ADJUST or BF_RECT);
		TextOut(DC, 110, LinePos +  2, Field.WinVer);
		TextOut(DC, 110, LinePos + 22, Field.Memory);

		RestoreDC(DC, -1);
		if wP = 0 then EndPaint(Wnd, PS);
		Result:= 0;
		Exit;
	end;
	WM_COMMAND:
		if (LongRec(wP).Hi = BN_CLICKED) and (HWND(lP) = Field.OkButton) then
		begin
			PostQuitMessage(1);
			Result := 0;
			Exit;
		end
		else if wP = IDCANCEL then
			PostQuitMessage(0);
	WM_CLOSE:
		PostQuitMessage(0);
{$IFDEF VERSION_ANIMATION}
	WM_TIMER:
		if wP = IDT_ANIMATE then
		begin
			if Field.AnimationMode <> amNone then
				InvalidateRect(Wnd, @PictureBorder, False);
			Result := 0;
			Exit;
		end;
{$ENDIF}
	end;

	Result := DefWindowProcW(Wnd, Msg, WP, LP);
end;

function VersionDialog(ParentWnd: HWND; const Title: WideString): Boolean;
const
	WndClassName = 'SVerDlg';
var
	WC: WNDCLASSEXW;
	VerWnd: HWND;
	M: MSG;
	Field: TVerWndField;
	V: TVersionBuffer;
	OS: string;
	MemStatus: MEMORYSTATUS;
{$IFDEF VERSION_ANIMATION}
	OrgRandSeed: LongInt;
{$ENDIF}
	SysMenu: HMENU;
	SR, AR: TRect;
begin
	if not GetVersionBuffer(GetModuleName(0), V) then
		Assert(False);

	//tB[h
	Field.V_Company   := GetVersionString(V, vrCompanyName);
	Field.V_Product   := GetVersionString(V, vrProductName);
	Field.V_Version   := GetVersionString(V, vrProductVersion);
	Field.V_Copyright := GetVersionString(V, vrLegalCopyright);
{$IFDEF VERSION_DESCRIPTION}
	Field.V_Comments  := GetVersionString(V, vrFileDescription);
{$ENDIF}
	case Win32Platform of
	VER_PLATFORM_WIN32_WINDOWS: OS:= 'Windows 9x';
	VER_PLATFORM_WIN32_NT:      OS:= 'Windows NT';
	else                        OS:= 'Windows';
	end;
	Field.WinVer:= Format('%s %d.%d {Build %d} %s', [OS, Win32MajorVersion,
					Win32MinorVersion, Win32BuildNumber, Win32CSDVersion]);
	MemStatus.dwLength:= SizeOf(MemStatus);
	GlobalMemoryStatus(MemStatus);
	Field.Memory := Format('Memory Available to Windows:%9.0n KB (Used %3d %%)',
					[MemStatus.dwTotalPhys / 1024, MemStatus.dwMemoryLoad]);
	Field.Icon:= LoadImageW(MainInstance, 'MAINICON', IMAGE_ICON, 32, 32, 0);

{$IFDEF VERSION_ANIMATION}
	OrgRandSeed:= RandSeed;
	Randomize;
	Field.AnimationMode:= TAnimationMode(
					Random(Ord(High(TAnimationMode)) - Ord(Low(TAnimationMode)) + 1));
	RandSeed:= OrgRandSeed;
{$ENDIF}

	//NX
	WC.cbSize:= SizeOf(WC);
	if not GetClassInfoExW(SysInit.hInstance, WndClassName, WC) then
	begin
		WC.style        := CS_DBLCLKS or CS_SAVEBITS or CS_BYTEALIGNWINDOW;
		WC.lpfnWndProc  := @VerWndProc;
		WC.cbClsExtra   := 0;
		WC.cbWndExtra   := 0;
		WC.hInstance    := MainInstance;
		WC.hIcon        := 0;
		WC.hCursor      := LoadCursorW(0, PWideChar(IDC_ARROW));
		WC.hbrBackground:= GetSysColorBrush(COLOR_3DFACE);
		WC.lpszMenuName := nil;
		WC.lpszClassName:= WndClassName;
		WC.hIconSm      := 0;
		{$IFOPT C+} Assert( {$ENDIF} RegisterClassExW(WC)	{$IFOPT C+}	<> 0); {$ENDIF}
	end;

	if ParentWnd <> 0 then
		EnableWindow(ParentWnd, False);

	{TCY}
	AR.Left := 0;
	AR.Top := 0;
	AR.Right := DlgWidth;
	AR.Bottom := DlgHeight;
	AdjustWindowRectEx(AR, DlgStyle, False, DlgExStyle);

	{ʃTCY}
	//SystemParametersInfoW(SPI_GETWORKAREA, 0, @SR, 0);
	SR := GetActiveMonitorRect;

	{쐬}
	VerWnd := CreateWindowExW(DlgExStyle, WndClassName, PWideChar(Title), DlgStyle,
					((SR.Right  - SR.Left) - (AR.Right - AR.Left)) div 2 + SR.Left,
					((SR.Bottom - SR.Top ) - (AR.Bottom - AR.Top)) div 2 + SR.Top,
					AR.Right - AR.Left, AR.Bottom - AR.Top,
					ParentWnd, 0, MainInstance, nil);

	Assert(VerWnd <> 0);

	SetWindowLongW(VerWnd, GWL_USERDATA, LongWord(@Field));

	{VXej[}
	SysMenu := GetSystemMenu(VerWnd, False);
	DeleteMenu(SysMenu, SC_TASKLIST, MF_BYCOMMAND);
	DeleteMenu(SysMenu, 7, MF_BYPOSITION);
	DeleteMenu(SysMenu, 5, MF_BYPOSITION);
	DeleteMenu(SysMenu, SC_MAXIMIZE, MF_BYCOMMAND);
	DeleteMenu(SysMenu, SC_MINIMIZE, MF_BYCOMMAND);
	DeleteMenu(SysMenu, SC_SIZE, MF_BYCOMMAND);
	DeleteMenu(SysMenu, SC_RESTORE, MF_BYCOMMAND);

	//{^
	Field.OkButton:= CreateWindowExW(0, 'BUTTON', 'OK',
					WS_VISIBLE or WS_CHILD or WS_TABSTOP or BS_DEFPUSHBUTTON,
					DlgWidth - 80, DlgHeight - 30, 75, 25, VerWnd, 0, hInstance, Nil);
	SendMessageW(Field.OkButton, WM_SETFONT, GetStockObject(DEFAULT_GUI_FONT), 0);

{$IFDEF VERSION_ANIMATION}
	//^C}[
	SetTimer(VerWnd, IDT_ANIMATE, 100, nil);
{$ENDIF}

	//\
	ShowWindow(VerWnd, SW_SHOW);
	SetFocus(Field.OkButton);

	{[v}
	while GetMessageW(M, 0, 0, 0) do
	begin
		if not IsDialogMessageW(VerWnd, M) then
		begin
			TranslateMessage(M);
			DispatchMessageW(M);
		end;
	end;

	if ParentWnd <> 0 then
	begin
		EnableWindow(ParentWnd, True);
		SetActiveWindow(ParentWnd);
	end;

	{j}
{$IFDEF VERSION_ANIMATION}
	KillTimer(VerWnd, IDT_ANIMATE);
{$ENDIF}
	DestroyIcon(Field.Icon);
	DestroyWindow(VerWnd);

	Result := M.wParam <> 0;
end;

{ Screen Utilities }

function SystemWorkArea: TRect;
begin
	SystemParametersInfoW(SPI_GETWORKAREA, 0, @Result, 0);
end;

function GetFullScreenRect: TRect;
begin
	Result.Left := Multimon.GetSystemMetrics(SM_XVIRTUALSCREEN);
	Result.Top := Multimon.GetSystemMetrics(SM_YVIRTUALSCREEN);
	Result.Right := Result.Left + Multimon.GetSystemMetrics(SM_CXVIRTUALSCREEN);
	Result.Bottom := Result.Top + Multimon.GetSystemMetrics(SM_CYVIRTUALSCREEN);
end;

function GetActiveMonitorRect(Window: HWND = 0): TRect;
var
	Monitor: THandle;
	Info: TMonitorInfoW;
begin
	if (Window = 0) or not IsWindowVisible(Window) then Window := GetActiveWindow;
	if (Window = 0) or not IsWindowVisible(Window) then Window := GetForegroundWindow;
	Monitor := MonitorFromWindow(Window, MONITOR_DEFAULTTONEAREST);
	Info.cbSize := SizeOf(Info);
	GetMonitorInfoW(Monitor, @Info);
	Result := Info.rcWork
end;	

{ Window Utilities }

function EnumWindowsProc(Wnd: HWND; var D: TMethod): LongBool; stdcall;
asm
	mov EDX, D
	mov EAX, TMethod[EDX].Data
	push EAX
	mov EAX, Wnd
	push EAX
	mov ECX, TMethod[EDX].Code
	call ECX
	add esp, 8
	xor al, 1
	cbw
	cwde
end;

procedure EnumerateWindows(Proc: CEachWindow; Context: Pointer);
var
	D: TMethod;
begin
	D.Code := @Proc;
	D.Data := Context;
	EnumWindows(@EnumWindowsProc, LPARAM(@D))
end;

procedure EnumerateChildWindows(Parent: HWND; Proc: CEachWindow; Context: Pointer);
var
	D: TMethod;
begin
	D.Code := @Proc;
	D.Data := Context;
	EnumChildWindows(Parent, @EnumWindowsProc, LPARAM(@D))
end;

function TWindowEnumerator.GetCurrent: HWND;
begin
	Result := HWND(inherited GetCurrent)
end;

function TWindowCollection.GetEnumerator: TWindowEnumerator;
	procedure Call;
	begin
		EnumerateWindows(CEachWindow(@TWindowEnumerator.Each), Result);
	end;
begin
	Result := TWindowEnumerator(TWindowEnumerator.GetEnumerator);
	Call;
end;

class function TWindowCollection.List: TWindowCollection; 
begin
end;

function TChildWindowCollection.GetEnumerator: TWindowEnumerator;
	procedure Call;
	begin
		EnumerateChildWindows(Parent, CEachWindow(@TWindowEnumerator.Each), Result);
	end;
begin
	Result := TWindowEnumerator(TWindowEnumerator.GetEnumerator);
	Call;
end;

class function TChildWindowCollection.List(AParent: HWND): TChildWindowCollection;
begin
	Result.Parent := AParent
end;

{ Cursor Utilities }

procedure SetCursorWait;
begin
	SetCursor(LoadCursorW(0, PWideChar(IDC_WAIT)));
end;

procedure ResetCursor;
var
	P: TPoint;
	Handle: HWND;
	Code: LRESULT;
	MakeLong: LongRec;
begin
	GetCursorPos(P);
	Handle := WindowFromPoint(P);
	if (Handle <> 0) and
			(GetWindowThreadProcessId(Handle, nil) = GetCurrentThreadId) then
	begin
		Code := SendMessageW(Handle, WM_NCHITTEST, 0, LPARAM(PointToSmallPoint(P)));

		MakeLong.Lo := Code;
		MakeLong.Hi := WM_MOUSEMOVE;

		SendMessageW(Handle, WM_SETCURSOR, Handle, LPARAM(MakeLong));
	end;
end;

function DragDetect(const From, Current: TPoint): Boolean;
begin
	Result := (Abs(Current.X - From.X) > Windows.GetSystemMetrics(SM_CXDRAG)) or
		(Abs(Current.Y - From.Y) > Windows.GetSystemMetrics(SM_CYDRAG))
end;

{ Keyboard Utilities }

function KeyNameText(VirtualKey: Byte): WideString;
var
	S: array[0..127] of WideChar;
	E: DWORD;
	L: Integer;
begin
	E := 0;
	if VirtualKey in [VK_SPACE + 1 .. VK_NUMPAD0 - 1] then
		E := 1 shl 24;

	L := GetKeyNameTextW(MapVirtualKeyW(VirtualKey, 0) shl 16 or E, S, 127);
	SetString(Result, S, L);
end;

{ Text Utilities }

function GetFontSize(Font: HFONT; const Text: WideString): TSize;
var
	DC: HDC;
	Org: THandle;
begin
	DC := GetDC(0);
	Org := SelectObject(DC, Font);
	GetTextExtentPoint32W(DC, PWideChar(Text), Length(Text), Result);
	SelectObject(DC, Org);
	ReleaseDC(0, DC);
end;

function GetTextSize(DC: HDC; const Text: WideString): TSize; overload;
begin
	GetTextExtentPoint32W(DC, Pointer(Text), Length(Text), Result);
end;

function TextOut(DC: HDC; X, Y: Integer; const Text: WideString): Boolean;
begin
	Result:= Windows.TextOutW(DC, X, Y, Pointer(Text), Length(Text));
end;

function MinimizeText(const Text: WideString; DC: HDC; MaxLen: Integer): WideString;
var
	DotWidth: Integer;
begin
	Result := Text;
	if GetTextSize(DC, Result).CX > MaxLen then
	begin
		DotWidth := GetTextSize(DC, '...').CX;
		repeat
			Delete(Result, Length(Result), 1);
		until GetTextSize(DC, Result).CX + DotWidth <= MaxLen;
		Result := Result + '...'
	end;
end;

function MinimizeName(const Filename: AnsiString; DC: HDC; MaxLen: Integer): AnsiString;
var
	Drive: TFileName;
	Dir: TFileName;
	Name: TFileName;
	Root: Boolean;
	P: Integer;
begin
	Result := SysUtils.StringReplace(FileName, '/', '\', [rfReplaceAll]);

	if (Length(Dir) >= 3) and (Dir[1] <> ':') then
	begin
		Dir := SysUtils.ExtractFilePath(Result);
		Name := SysUtils.ExtractFileName(Result);
	end
	else
	begin
		P := SysUtils.LastDelimiter('\', FileName);
		Dir := Copy(Result, 1, P);
		Name := Copy(Result, P + 1, MaxInt);
	end;

	if (Length(Dir) >= 2) and (Dir[1] <> ':') and (Dir[2] = ':') then
	begin
		Drive := Copy(Dir, 1, 2);
		Delete(Dir, 1, 2);
	end
	else
		Drive := '';

	while ((Dir <> '') or (Drive <> '')) and
					(GetTextSize(DC, Result).CX > MaxLen) do
	begin
		if Dir = '\...\' then
		begin
			Drive := '';
			Dir := '...\';
		end
		else if Dir = '' then
			Drive := ''
		else if Dir = '\' then
			Dir := ''
		else
		begin
			if Dir[1] = '\' then
			begin
				Root := True;
				Delete(Dir, 1, 1);
			end
			else
				Root := False;
			if Dir[1] = '.' then
				Delete(Dir, 1, 4);
			P := AnsiPos('\',Dir);
			if P <> 0 then
			begin
				Delete(Dir, 1, P);
				Dir := '...\' + Dir;
			end
			else
				Dir := '';
			if Root then
				Dir := '\' + Dir;
		end;

		Result := Drive + Dir + Name;
	end;

	if GetTextSize(DC, Result).CX > MaxLen then
	begin
		Result := Result + '...';
		repeat
			Delete(Result, Length(Result) - 3, 1);
		until (Length(Result) <= 4) or (GetTextSize(DC, Result).CX <= MaxLen)
	end
end;

function MinimizeName(const Filename: WideString; DC: HDC; MaxLen: Integer): WideString;
var
	Drive: WideString;
	Dir: WideString;
	Name: WideString;
	Root: Boolean;
	I, P: Integer;
begin
	Result := FileName;
	for I := 1 to Length(Result) do
		if Result[I] = '/' then
			Result[I] := '\';

	if (Length(Dir) >= 3) and (Dir[1] <> ':') then
	begin
		Dir := SysUtilsW.ExtractFilePath(Result);
		Name := SysUtilsW.ExtractFileName(Result);
	end
	else
	begin
		P := WideLastDelimiter('\', FileName);
		Dir := Copy(Result, 1, P);
		Name := Copy(Result, P + 1, MaxInt);
	end;

	if (Length(Dir) >= 2) and (Dir[1] <> ':') and (Dir[2] = ':') then
	begin
		Drive := Copy(Dir, 1, 2);
		Delete(Dir, 1, 2);
	end
	else
		Drive := '';

	while ((Dir <> '') or (Drive <> '')) and
					(GetTextSize(DC, Result).CX > MaxLen) do
	begin
		if Dir = '\...\' then
		begin
			Drive := '';
			Dir := '...\';
		end
		else if Dir = '' then
			Drive := ''
		else if Dir = '\' then
			Dir := ''
		else
		begin
			if Dir[1] = '\' then
			begin
				Root := True;
				Delete(Dir, 1, 1);
			end
			else
				Root := False;
			if Dir[1] = '.' then
				Delete(Dir, 1, 4);
			P := Pos('\', Dir);
			if P <> 0 then
			begin
				Delete(Dir, 1, P);
				Dir := '...\' + Dir;
			end
			else
				Dir := '';
			if Root then
				Dir := '\' + Dir;
		end;

		Result := Drive + Dir + Name;
	end;

	if GetTextSize(DC, Result).CX > MaxLen then
	begin
		Result := Result + '...';
		repeat
			Delete(Result, Length(Result) - 3, 1);
		until (Length(Result) <= 4) or (GetTextSize(DC, Result).CX <= MaxLen)
	end
end;

{ Flat Scrollbar Utilities }

function _GetScrollProp(p1: HWND; propIndex: Integer; p3: PInteger): Bool; stdcall;
begin
	Result := True
end;

function _SetScrollProp(p1: HWND; index: Integer; newValue: Integer; p4: Bool): Bool; stdcall;
begin
	Result := True
end;

function _InitializeSB(Wnd: HWND): Bool; stdcall;
begin
	Result := True
end;

procedure _UninitializeSB(Wnd: HWND); stdcall;
begin
end;

var
	_FlatScrollBarManager: TScrollBarManager;
	FlatScrollBarChecked: Boolean;

function FlatScrollBarManager: PScrollBarManager;
label
	Good, Bad;
const
	cctrl = 'comctl32.dll';
var
	ComCtl32DLL: HMODULE;
	P: Pointer;
begin
	if not FlatScrollBarChecked then
	begin
		FlatScrollBarChecked := True;
		ComCtl32DLL := GetModuleHandleW(cctrl);

		P := GetProcAddress(ComCtl32DLL, 'FlatSB_EnableScrollBar');
		if P = nil then goto Bad;
		_FlatScrollBarManager.EnableScrollBar := P;

		P := GetProcAddress(ComCtl32DLL, 'FlatSB_ShowScrollBar');
		if P = nil then goto Bad;
		_FlatScrollBarManager.ShowScrollBar := P;

		P := GetProcAddress(ComCtl32DLL, 'FlatSB_GetScrollInfo');
		if P = nil then goto Bad;
		_FlatScrollBarManager.GetScrollInfo := P;

		P := GetProcAddress(ComCtl32DLL, 'FlatSB_SetScrollInfo');
		if P = nil then goto Bad;
		_FlatScrollBarManager.SetScrollInfo := P;

		P := GetProcAddress(ComCtl32DLL, 'FlatSB_GetScrollRange');
		if P = nil then goto Bad;
		_FlatScrollBarManager.GetScrollRange := P;

		P := GetProcAddress(ComCtl32DLL, 'FlatSB_SetScrollRange');
		if P = nil then goto Bad;
		_FlatScrollBarManager.SetScrollRange := P;

		P := GetProcAddress(ComCtl32DLL, 'FlatSB_GetScrollPos');
		if P = nil then goto Bad;
		_FlatScrollBarManager.GetScrollPos := P;

		P := GetProcAddress(ComCtl32DLL, 'FlatSB_SetScrollPos');
		if P = nil then goto Bad;
		_FlatScrollBarManager.SetScrollPos := P;

		P := GetProcAddress(ComCtl32DLL, 'FlatSB_GetScrollProp');
		if P = nil then goto Bad;
		_FlatScrollBarManager.GetScrollProp := P;

		P := GetProcAddress(ComCtl32DLL, 'FlatSB_SetScrollProp');
		if P = nil then goto Bad;
		_FlatScrollBarManager.SetScrollProp := P;

		P := GetProcAddress(ComCtl32DLL, 'InitializeFlatSB');
		if P = nil then goto Bad;
		_FlatScrollBarManager.Initialize := P;

		P := GetProcAddress(ComCtl32DLL, 'UninitializeFlatSB');
		if P = nil then goto Bad;
		_FlatScrollBarManager.Uninitialize := P;

		goto Good;
Bad:
		_FlatScrollBarManager := _DefaultScrollBarManager;
Good:
	end;
	Result := @_FlatScrollBarManager;
end;

end.

