unit SysUtilsW;

{*******************************************************}
{                                                       }
{       Delphi Run-time Library                         }
{       Wide-Version Functions of SysUtils.pas          }
{                                                       }
{       2001..2007 YT                                   }
{                                                       }
{*******************************************************}

interface

uses
	Windows, Types, SysUtils;

{ API }

function GetLongPathNameA(ShortPath: PAnsiChar; LongPath: PAnsiChar; Buffer: DWORD): DWORD; stdcall; external kernel32;
function GetLongPathNameW(ShortPath: PWideChar; LongPath: PWideChar; Buffer: DWORD): DWORD; stdcall; external kernel32;

{ Wide version functions of SysUtils }

type
	TAnsiFileName = TFileName;
	TWideFileName = WideString;

function ParamStr(I: Integer): WideString;
function ParamStrings(CmdLine: PWideChar): TWideStringDynArray;

function WideStrScan(Str: PWideChar; C: WideChar): PWideChar; register;
function WideStrLScan(Str: PWideChar; Length: Integer; C: WideChar): PWideChar; register;
function WideStrRScan(Str: PWideChar; C: WideChar): PWideChar; register;
function WideStrEnd(const Str: PWideChar): PWideChar; register;
function WideStrLComp(S1, S2: PWideChar; Len: Integer): Integer; register;
function WideStrLIComp(S1, S2: PWideChar; Len: Integer): Integer; register;
function WideUpCase(C: WideChar): WideChar; register;
function WideLoCase(C: WideChar): WideChar; register;
function WideStrTrimLeft(S: PWideChar): PWideChar; register;

function WideStringReplace(const S, OldPattern, NewPattern: WideString; Flags: TReplaceFlags): WideString;
function WideExtractQuotedStr(var Src: PWideChar; Quote: WideChar): WideString;
function WideQuotedStr(const S: WideString; Q: WideChar): WideString;
function WideLastDelimiter(const Delimiters, S: WideString): Integer;

function CompareFileName(S1: PWideChar; L1: Cardinal; S2: PWideChar; L2: Cardinal): Integer; overload;
function CompareFileName(const S1, S2: WideString): Integer; overload;
function ExtractFileDir(const FileName: WideString): WideString;
function ExtractFilePath(const FileName: WideString): WideString;
function ExtractFileDrive(const FileName: WideString): WideString;
function ExtractFileExt(const FileName: WideString): WideString;
function ExtractFileName(const FileName: WideString): WideString;
function ExpandFileName(const FileName: WideString): WideString;
function ChangeFileExt(const FileName, Extension: WideString): WideString;
function IncludeTrailingPathDelimiter(const S: WideString): WideString;
function ExcludeTrailingPathDelimiter(const S: WideString): WideString;
function ExtractShortPathName(const FileName: WideString): WideString;
function ExtractLongPathName(const FileName: WideString): WideString;
function ExtractRelativePath(const Path, FileName: WideString): WideString;

function FileOpen(const FileName: WideString; Mode: LongWord): Integer;
function FileCreate(const FileName: WideString): Integer;

function FileExists(const FileName: WideString): Boolean;
function DirectoryExists(const Directory: WideString): Boolean;
function GetCurrentDir: WideString;
procedure SetCurrentDir(const Dir: WideString);

function GetEnvironmentVariable(const Name: WideString): WideString; overload;

{ Binary }

type
	BinaryString = type AnsiString;

function BinaryToUnicode(const S: BinaryString): WideString;
function UnicodeToBinary(const S: WideString): BinaryString;
function CompareBinary(X: Pointer; Y: Pointer; Size: Integer): Integer; overload;
function CompareBinary(S1: Pointer; L1: Integer; S2: Pointer; L2: Integer): Integer; overload;

{ Ansi }

function AnsiByteToCharLen(S: PChar; Len: Integer; MaxLen: Integer): Integer;
function AnsiCharToByteLen(S: PChar; Len: Integer; MaxLen: Integer): Integer;
function LoCase(C: Char): Char; register;
function StrTrimLeft(S: PChar): PChar; register;
function StrLRScan(Str: PChar; Chr: Char; Len: Integer): PChar; register;

{ UTF-8 }

const
	UTF8Lead2Bytes = [#$c0..#$df];
	UTF8Lead3Bytes = [#$e0..#$ef];

function UTF8CharToByteLen(const S: UTF8String; MaxLen: Integer): Integer; overload;
function UTF8CharToByteLen(S: PChar; Len: Integer; MaxLen: Integer): Integer; overload;
function UTF8ByteToCharLen(const S: UTF8String; MaxLen: Integer): Integer; overload;
function UTF8ByteToCharLen(S: PChar; Len: Integer; MaxLen: Integer): Integer; overload;

{ EUC-JP }

(*
type
	EUCString = type AnsiString;

const
	EUCLeadBytes = [#$8e, #$a1..#$fe];

function EUCJPToAnsi(const S: EUCString): AnsiString;
function AnsiToEUCJP(const S: AnsiString): EUCString;
function EUCCharToByteLen(const S: EUCString; MaxLen: Integer): Integer; overload;
function EUCCharToByteLen(S: PChar; Len: Integer; MaxLen: Integer): Integer; overload;
function EUCByteToCharLen(const S: EUCString; MaxLen: Integer): Integer; overload;
function EUCByteToCharLen(S: PChar; Len: Integer; MaxLen: Integer): Integer; overload;

function ShiftJisToJis(Code: Word): Word; register;
function JisToShiftJis(Code: Word): Word; register;
function EucJpToJis(Code: Word): Word; register;
function JisToEucJp(Code: Word): Word; register;
*)

implementation

function GetParamStr(P: PWideChar; var S: PWideChar; var Len: Integer): PWideChar;
begin
	Result := P;
	while Result^ <= #32 do
	begin
		if Result^ = #0 then
		begin
			S := nil;
			Len := 0;
			Exit;
		end;
		Inc(Result);
	end;
	if Result^ = '"' then
	begin
		Inc(Result);		
		S := Result;
		while (Result^ <> '"') and (Result^ <> #0) do Inc(Result);
		Len := Result - S;
		if Result^ <> #0 then Inc(Result);
	end
	else
	begin
		S := Result;
		repeat Inc(Result) until Result^ <= #32;
		Len := Result - S;
	end;
end;

function ParamStr(I: Integer): WideString;
var
	P, S: PWideChar;
	Len: Integer;
begin
	P := GetCommandLineW;
	repeat
		P := GetParamStr(P, S, Len);
		Dec(I);
	until I < 0;
	SetString(Result, S, Len)
end;

function ParamStrings(CmdLine: PWideChar): TWideStringDynArray;
var
	P, S: PWideChar;
	Index, Len: Integer;
begin
	Result := nil;
	P := CmdLine;
	repeat
		P := GetParamStr(P, S, Len);
		if Len = 0 then Break;
		Index := Length(Result);
		SetLength(Result, Index + 1);
		SetString(Result[Index], S, Len)
	until False;
end;

function WideStrScan(Str: PWideChar; C: WideChar): PWideChar; register;
asm
	PUSH    EDI
	PUSH    EAX
	MOV     EDI, eax{Str}
	MOV     ECX,$FFFFFFFF
	XOR     ax,ax
	REPNE   scasw
	NOT     ECX
	POP     EDI
	MOV     ax, dx{Chr}
	REPNE   scasw
	MOV     EAX,0
	JNE     @Exit
	MOV     EAX,EDI
	sub     EAX, 2
@Exit:
	POP     EDI
end;

function WideStrLScan(Str: PWideChar; Length: Integer; C: WideChar): PWideChar; register;
asm
	push edi
	mov edi, eax
	mov ax, cx
	mov ecx, edx
	repne scasw
	mov eax, edi
	je @Exit
	mov eax, 2
@Exit:
	sub eax, 2
	pop edi
end;

function WideStrRScan(Str: PWideChar; C: WideChar): PWideChar;
var
	S: PWideChar;
begin
	Result := nil;
	S := Str;
	while S^ <> #0 do
	begin
		if S^ = C then Result := S;
		Inc(S);
	end;
end;

function WideStrEnd(const Str: PWideChar): PWideChar; register;
asm
	MOV     EDX,EDI
	MOV     EDI,EAX
	MOV     ECX,0FFFFFFFFH
	XOR     AX,AX
	REPNE   SCASW
	LEA     EAX,[EDI-2]
	MOV     EDI,EDX
end;

function WideStrLComp(S1, S2: PWideChar; Len: Integer): Integer; register;
asm
	PUSH EDI
	PUSH ESI
	PUSH EBX
	MOV ESI, EAX {S1}
	MOV EDI, EDX {S2}
	MOV EBX, ECX {Len}
	xor eax, eax
	test ecx, ecx
	je @0
	REPNE SCASW
	SUB EBX, ECX //Len  WideStrLen(S2) ̒Zebxɓ
	MOV ECX, EBX
	MOV EDI, EDX
	REPE CMPSW
	je @0
	ja @1
	dec eax
	jmp @0
@1:
	inc eax
@0:
	POP EBX
	POP ESI
	POP EDI
end;

function WideStrLIComp(S1, S2: PWideChar; Len: Integer): Integer;
begin
	Result := CompareStringW(LOCALE_USER_DEFAULT, NORM_IGNORECASE, S1, Len, S2, Len) - 2
end;

function WideUpCase(C: WideChar): WideChar; register;
asm
	cmp AX, 'a'; jb @EndProc;
	cmp AX, 'z'; ja @EndProc;
	sub AX, $20;
@EndProc:
end;

function WideLoCase(C: WideChar): WideChar; register;
asm
	cmp AX, 'A'; jb @EndProc;
	cmp AX, 'Z'; ja @EndProc;
	add AX, $20;
@EndProc:
end;

function WideStrTrimLeft(S: PWideChar): PWideChar; register;
begin
	Result := S;
	while (Result^ <> #0) and (Result^ <= #$20) do
		Inc(Result);
end;

function WideStringReplace(const S, OldPattern, NewPattern: WideString; Flags: TReplaceFlags): WideString;
var
	SearchStr, Patt, NewStr: WideString;
	Offset: Integer;
begin
	if rfIgnoreCase in Flags then
	begin
		SearchStr := WideUpperCase(S);
		Patt := WideUpperCase(OldPattern);
	end else
	begin
		SearchStr := S;
		Patt := OldPattern;
	end;
	NewStr := S;
	Result := '';
	while SearchStr <> '' do
	begin
		Offset := Pos(Patt, SearchStr);
		if Offset = 0 then
		begin
			Result := Result + NewStr;
			Break;
		end;
		Result := Result + Copy(NewStr, 1, Offset - 1) + NewPattern;
		NewStr := Copy(NewStr, Offset + Length(OldPattern), MaxInt);
		if not (rfReplaceAll in Flags) then
		begin
			Result := Result + NewStr;
			Break;
		end;
		SearchStr := Copy(SearchStr, Offset + Length(Patt), MaxInt);
	end;
end;

function WideExtractQuotedStr(var Src: PWideChar; Quote: WideChar): WideString;
var
	P, Dest: PWideChar;
	DropCount: Integer;
begin
	Result := '';
	if (Src = nil) or (Src^ <> Quote) then Exit;
	Inc(Src);
	DropCount := 1;
	P := Src;
	Src := WideStrScan(Src, Quote);
	while Src <> nil do   // count adjacent pairs of quote chars
	begin
		Inc(Src);
		if Src^ <> Quote then Break;
		Inc(Src);
		Inc(DropCount);
		Src := WideStrScan(Src, Quote);
	end;
	if Src = nil then Src := WideStrEnd(P);
	if ((Src - P) <= 1) then Exit;
	if DropCount = 1 then
		SetString(Result, P, Src - P - 1)
	else
	begin
		SetLength(Result, Src - P - DropCount);
		Dest := PWideChar(Result);
		Src := WideStrScan(P, Quote);
		while Src <> nil do
		begin
			Inc(Src);
			if Src^ <> Quote then Break;
			Move(P^, Dest^, Src - P);
			Inc(Dest, Src - P);
			Inc(Src);
			P := Src;
			Src := WideStrScan(Src, Quote);
		end;
		if Src = nil then Src := WideStrEnd(P);
		Move(P^, Dest^, Src - P - 1);
	end;
end;

function WideQuotedStr(const S: WideString; Q: WideChar): WideString;
var
	Q2: WideString;
begin
	SetLength(Q2, 2);
	Q2[1] := Q;
	Q2[2] := Q;
	Result := Q + WideStringReplace(S, Q, Q2, [rfReplaceAll]) + Q;
end;

function WideLastDelimiter(const Delimiters, S: WideString): Integer;
begin
	Result := Length(S);
	while Result > 0 do
	begin
		if WideStrLScan(Pointer(Delimiters), Length(Delimiters), S[Result]) <> nil then Break;
		Dec(Result);
	end;
end;

function CompareFileName(S1: PWideChar; L1: Cardinal; S2: PWideChar; L2: Cardinal): Integer; overload;
begin
	Result := CompareStringW(LOCALE_SYSTEM_DEFAULT, NORM_IGNORECASE, S1, L1, S2, L2) - 2
end;

function CompareFileName(const S1, S2: TWideFileName): Integer; overload;
begin
	Result := CompareStringW(LOCALE_SYSTEM_DEFAULT, NORM_IGNORECASE,
		PWideChar(S1), Length(S1), PWideChar(S2), Length(S2)) - 2;
end;

function ExtractFileDir(const FileName: TWideFileName): TWideFileName;
var
	I: Integer;
begin
	I := WideLastDelimiter(PathDelim + DriveDelim, FileName);
	if (I > 1) and (FileName[I] = PathDelim) and (((FileName[I - 1] <> PathDelim)
		and (FileName[I-1] <> DriveDelim))) then
	begin
		Dec(I);
	end;
	Result := Copy(FileName, 1, I);
end;

function ExtractFilePath(const FileName: TWideFileName): TWideFileName;
var
	I: Integer;
begin
	I := WideLastDelimiter(PathDelim + DriveDelim, FileName);
	Result := Copy(FileName, 1, I);
end;

function ExtractFileDrive(const FileName: TWideFileName): TWideFileName;
var
	I, J: Integer;
begin
	if (Length(FileName) >= 2) and (FileName[2] = DriveDelim) then
		Result := Copy(FileName, 1, 2)
	else if (Length(FileName) >= 2) and (FileName[1] = PathDelim) and
					(FileName[2] = PathDelim) then
	begin
		J := 0;
		I := 3;
		while (I < Length(FileName)) and (J < 2) do
		begin
			if FileName[I] = PathDelim then Inc(J);
			if J < 2 then Inc(I);
		end;
		if FileName[I] = PathDelim then Dec(I);
		Result := Copy(FileName, 1, I);
	end
	else
		Result := '';
end;

function ExtractFileExt(const FileName: TWideFileName): TWideFileName;
var
	I: Integer;
begin
	I := WideLastDelimiter('.' + PathDelim + DriveDelim, FileName);
	if (I > 0) and (FileName[I] = '.') then
		Result := Copy(FileName, I, MaxInt)
	else
		Result := '';
end;

function ExtractFileName(const FileName: TWideFileName): TWideFileName;
var
	I: Integer;
begin
	I := WideLastDelimiter(PathDelim + DriveDelim, FileName);
	Result := Copy(FileName, I + 1, MaxInt);
end;

function ExpandFileName(const FileName: TWideFileName): TWideFileName;
var
	FName: PWideChar;
	Buffer: array[0..MAX_PATH - 1] of WideChar;
begin
	SetString(Result, Buffer, GetFullPathNameW(PWideChar(FileName), Length(Buffer), Buffer, FName));
end;

function ChangeFileExt(const FileName, Extension: WideString): WideString;
var
	I: Integer;
begin
	I := WideLastDelimiter('.' + PathDelim + DriveDelim, Filename);
	if (I = 0) or (FileName[I] <> '.') then I := MaxInt;
	Result := Copy(FileName, 1, I - 1) + Extension;
end;

function IncludeTrailingPathDelimiter(const S: TWideFileName): TWideFileName;
begin
	Result := S;
	if (Result = '') or (Result[Length(Result)] <> PathDelim) then begin
		Result := Result + PathDelim;
	end
end;

function ExcludeTrailingPathDelimiter(const S: WideString): WideString;
begin
	Result := S;
	if (Result <> '') and (Result[Length(Result)] = PathDelim) then begin
		Delete(Result, Length(Result), 1);
	end
end;

function ExtractShortPathName(const FileName: WideString): WideString;
const
	BufLen = MAX_PATH;
var
	L: DWORD;
begin
	SetLength(Result, BufLen);
	L := GetShortPathNameW(PWideChar(FileName), PWideChar(Result), BufLen + 1);
	SetLength(Result, L);
	if L > BufLen then
	begin
		SetLength(Result, GetShortPathNameW(PWideChar(FileName), PWideChar(Result), L));
  end	
end;

function ExtractLongPathName(const FileName: WideString): WideString;
const
	BufLen = MAX_PATH;
var
	L: DWORD;
begin
	SetLength(Result, BufLen);
	L := GetLongPathNameW(PWideChar(FileName), PWideChar(Result), BufLen + 1);
	SetLength(Result, L);
	if L > BufLen then
	begin
		SetLength(Result, GetShortPathNameW(PWideChar(FileName), PWideChar(Result), L));
  end	
end;

function ExtractRelativePath(const Path, FileName: WideString): WideString;
var
	Path2: WideString;
	L, I, Up: Integer;
begin
	if ExtractFileDrive(Path) = ExtractFileDrive(FileName) then
	begin
		Path2 := ExtractFileDir(Path);
		Result := FileName;
		Up := 0;
		while (Length(FileName) <= Length(Path2)) or
			(FileName[Length(Path2) + 1] <> PathDelim) or
			(CompareFileName(Pointer(Path2), Length(Path2), Pointer(FileName), Length(Path2)) <> 0) do
		begin
			L := Length(Path2);
			Path2 := ExtractFileDir(Path2);
			if Length(Path2) = L then Break;
			Inc(Up);
		end;
		Delete(Result, 1, Length(Path2));
		if Result[1] = '\' then
		begin
			Delete(Result, 1, 1);
		end;
		for I := Up downto 1 do
		begin
			Result := '..\' + Result;
		end
	end
	else
	begin
		Result := FileName
	end
end;

function FileOpen(const FileName: WideString; Mode: LongWord): Integer;
const
	AccessMode: array[0..2] of LongWord = (
		GENERIC_READ,
		GENERIC_WRITE,
		GENERIC_READ or GENERIC_WRITE);
	ShareMode: array[0..4] of LongWord = (
		0,
		0,
		FILE_SHARE_READ,
		FILE_SHARE_WRITE,
		FILE_SHARE_READ or FILE_SHARE_WRITE);
begin
	Result := -1;
	if ((Mode and 3) <= fmOpenReadWrite) and
		(((Mode and $F0) shr 4) <= fmShareDenyNone) then
		Result := Integer(CreateFileW(PWideChar(FileName), AccessMode[Mode and 3],
			ShareMode[(Mode and $F0) shr 4], nil, OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL, 0));
end;

function FileCreate(const FileName: WideString): Integer; overload;
begin
	Result := Integer(CreateFileW(PWideChar(FileName), GENERIC_READ or GENERIC_WRITE,
		0, nil, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0));
end;

function FileExists(const FileName: WideString): Boolean;
var
	Code: DWORD;
begin
	Code := GetFileAttributesW(PWideChar(FileName));
	Result := (Integer(Code) <> -1) and (FILE_ATTRIBUTE_DIRECTORY and Code = 0)
end;

function DirectoryExists(const Directory: WideString): Boolean;
var
	Code: DWORD;
begin
	Code := GetFileAttributesW(PWideChar(Directory));
	Result := (Integer(Code) <> -1) and (FILE_ATTRIBUTE_DIRECTORY and Code <> 0)
end;

function GetCurrentDir: WideString;
var
	P: array[0..MAX_PATH] of WideChar;
begin
	SetString(Result, P, GetCurrentDirectoryW(MAX_PATH, P));
end;

procedure SetCurrentDir(const Dir: WideString);
begin
  SetCurrentDirectoryW(PWideChar(Dir))
end;

function GetEnvironmentVariable(const Name: WideString): WideString;
var
	Len: integer;
begin
	Result := '';
	Len := Windows.GetEnvironmentVariableW(PWideChar(Name), nil, 0);
	if Len > 0 then
	begin
		SetLength(Result, Len - 1);
		Windows.GetEnvironmentVariableW(PWideChar(Name), PWideChar(Result), Len);
	end;
end;

{ Binary }

function BinaryToUnicode(const S: BinaryString): WideString;
begin
	SetString(Result, PWideChar(Pointer(S)), Length(S) div 2)
end;

function UnicodeToBinary(const S: WideString): BinaryString;
begin
	SetString(Result, PChar(Pointer(S)), Length(S) * 2)
end;

function CompareBinary(X : Pointer; Y : Pointer; Size: Integer): Integer;
asm
	push esi
	push edi
	mov esi, eax {X}
	mov edi, edx {Y}
	{mov ecx, Size}
	xor eax, eax
	repe cmpsb
	je @0
	ja @1
	dec eax
	jmp @0
@1:
	inc eax
@0:
	pop edi
	pop esi
end;

function CompareBinary(S1: Pointer; L1: Integer; S2: Pointer; L2: Integer): Integer;
asm
	PUSH ESI
	PUSH EDI
	MOV ESI, EAX {S1}
	MOV EDI, ECX {S2}
	mov eax, edx {L1}
	mov edx, L2
	mov ecx, eax
	sub eax, edx
	cmova ecx, edx
	xor edx, edx
	repe cmpsb
	je @0
	ja @1
	dec eax
	jmp @0
@1:
	inc eax
@0:
	POP EDI
	POP ESI
end;

{ Ansi }

function AnsiByteToCharLen(S: PChar; Len: Integer; MaxLen: Integer): Integer;
var
	I: Integer;
begin
	Result := 1;
	I := 0;
	while (I < MaxLen) and (I < Len) do
	begin
		if (S + I)^ in LeadBytes then Inc(I);
		Inc(I);
		Inc(Result);
	end;
end;

function AnsiCharToByteLen(S: PChar; Len: Integer; MaxLen: Integer): Integer;
var
	CI: Integer;
begin
	CI := 1;
	Result := 0;
	while (CI < MaxLen) and (Result < Len) do
	begin
		if (S + Result)^ in LeadBytes then Inc(Result);
		Inc(Result);
		Inc(CI);
	end;
end;

function LoCase(C: Char): Char;
asm
	cmp AL, 'A'; jb @EndProc;
	cmp AL, 'Z'; ja @EndProc;
	add AL, $20;
@EndProc:
end;

function StrTrimLeft(S: PChar): PChar;
asm
@While:
	mov dl, [eax]
	cmp dl, $00; je @EndProc;
	cmp dl, $20; ja @EndProc;
	inc EAX;
	jmp @While;
@EndProc:
end;

function StrLRScan(Str: PChar; Chr: Char; Len: Integer): PChar; register;
asm
	push ebx
	PUSH EDI
	MOV EDI, eax{Str}
	mov ebx, ecx
	XOR AL,AL
	REPNE SCASB
	STD
	DEC EDI
	neg ecx
	add ecx, ebx
	MOV     AL, dl{Chr}
	REPNE   SCASB
	MOV     EAX,0
	JNE     @@1
	MOV     EAX,EDI
	INC     EAX
@@1:
	CLD
	POP EDI
	pop ebx
end;

{ UTF-8 }

function UTF8ByteToCharLen(const S: UTF8String; MaxLen: Integer): Integer;
begin
	Result := UTF8ByteToCharLen(PChar(S), Length(S), MaxLen)
end;

function UTF8ByteToCharLen(S: PChar; Len: Integer; MaxLen: Integer): Integer;
var
	I: Integer;
begin
	Result := 1;
	I := 0;
	while (I < MaxLen) and (I < Len) do
	begin
		if (S + I)^ in UTF8Lead2Bytes then
			Inc(I, 2)
		else if (S + I)^ in UTF8Lead3Bytes then
			Inc(I, 3)
		else
			Inc(I);

		Inc(Result);
	end;
end;

function UTF8CharToByteLen(const S: UTF8String; MaxLen: Integer): Integer;
begin
	Result := UTF8CharToByteLen(PChar(S), Length(S), MaxLen)
end;

function UTF8CharToByteLen(S: PChar; Len: Integer; MaxLen: Integer): Integer;
var
	CI: Integer;
begin
	CI := 1;
	Result := 0;
	while (CI < MaxLen) and (Result < Len) do
	begin
		if (S + Result)^ in UTF8Lead2Bytes then
			Inc(Result, 2)
		else if (S + Result)^ in UTF8Lead3Bytes then
			Inc(Result, 3)
		else
			Inc(Result);

		Inc(CI);
	end;
end;

{ EUC-JP }

(*
function EUCJPToAnsi(const S: EUCString): AnsiString;
var
	I, Len: Integer;
	E: Word;
	PS, PD: PChar;
begin
	Len := Length(S);

	PS := Pointer(S);

	SetString(Result, PS, Len); {͕ςȂ͂}
	PD := Pointer(Result);

	I := 0;
	while I < Len do
	begin
		if (PS[I] in EUCLeadBytes) and (I + 1 < Len) then
		begin
			WordRec(E).Hi := Byte(PS[I]);
			WordRec(E).Lo := Byte(PS[I + 1]);

			E := JISToShiftJIS(EUCJPToJIS(E));

			Byte(PD[I])     := WordRec(E).Hi;
			Byte(PD[I + 1]) := WordRec(E).Lo;

			Inc(I);
		end;
		Inc(I);
	end;
end;

function AnsiToEUCJP(const S: AnsiString): EUCString;
var
	I, Len: Integer;
	E: Word;
	PS, PD: PChar;
begin
	Len := Length(S);

	PS := Pointer(S);

	SetString(Result, PS, Len); {͕ςȂ͂}
	PD := Pointer(Result);

	I := 0;
	while I < Len do
	begin
		if (PS[I] in LeadBytes) and (I + 1 < Len) then
		begin
			WordRec(E).Hi := Byte(PS[I]);
			WordRec(E).Lo := Byte(PS[I + 1]);

			E := JISToEUCJP(ShiftJISToJIS(E));

			Byte(PD[I])     := WordRec(E).Hi;
			Byte(PD[I + 1]) := WordRec(E).Lo;

			Inc(I);
		end;
		Inc(I);
	end;
end;

function EUCByteToCharLen(const S: EUCString; MaxLen: Integer): Integer;
begin
	Result := EUCByteToCharLen(PChar(S), Length(S), MaxLen)
end;

function EUCByteToCharLen(S: PChar; Len: Integer; MaxLen: Integer): Integer;
var
	I: Integer;
begin
	Result := 1;
	I := 0;
	while (I < MaxLen) and (I < Len) do
	begin
		if (S + I)^ in EUCLeadBytes then Inc(I);
		Inc(I);
		Inc(Result);
	end;
end;

function EUCCharToByteLen(const S: EUCString; MaxLen: Integer): Integer;
begin
	Result := EUCCharToByteLen(PChar(S), Length(S), MaxLen)
end;

function EUCCharToByteLen(S: PChar; Len: Integer; MaxLen: Integer): Integer;
var
	CI: Integer;
begin
	CI := 1;
	Result := 0;
	while (CI < MaxLen) and (Result < Len) do
	begin
		if (S + Result)^ in EUCLeadBytes then Inc(Result);
		Inc(Result);
		Inc(CI);
	end;
end;

function ShiftJisToJis(Code: Word): Word; register;
asm
	and ah, 3Fh
	shl ah, 1
	sub al, 9Fh
	jae @SKIP
	dec ah
	cmp al, 80h-9Fh
	adc al, 9Fh-40h-1
@SKIP:
	add ax, 2021h
end;

function JisToShiftJis(Code: Word): Word; register;
asm
	add ax, (9Fh-21h) + (42h-21h) * 256
	shr ah, 1
	jc @SKIP
	cmp al, 60h+(9Fh-21h)
	sbb al, 9Fh-40h-1
@SKIP:
	xor ah, 0A0h
end;

function EucJpToJis(Code: Word): Word; register;
asm
	sub ax, 8080h
end;

function JisToEucJp(Code: Word): Word; register;
asm
	add ax, 8080h
end;
*)

end.

